from pathlib import Path
from typing import Any

import typer
from pydantic import BaseModel, Field
import os

CONFIG_PATH_STR = os.getenv("LABELR_CONFIG_PATH", None)

CONFIG_PATH = (
    Path(CONFIG_PATH_STR).expanduser()
    if CONFIG_PATH_STR
    else Path("~").expanduser() / ".config/labelr/config.json"
)


def check_required_field(name: str, value: Any):
    """Check if a field if not null or empty, raise a BadParameter exception if it is."""
    if not value:
        raise typer.BadParameter(f"'{name}' is required.")


# validate_assignment allows to validate the model everytime it is updated
class LabelrConfig(BaseModel, validate_assignment=True):
    label_studio_url: str = Field(
        default="http://127.0.0.1:8080",
        description="URL of the Label Studio instance to use. Defaults to http://127.0.0.1:8080.",
    )
    label_studio_api_key: str | None = Field(
        default=None,
        description="API key for Label Studio.",
    )
    label_studio_project_id: int | None = Field(
        default=None, description="ID of the Label Studio project to use"
    )


def get_config() -> LabelrConfig:
    """Get labelr configuration.

    The configuration can come from (by order of precedence):
    - Environment variables
    - JSON file (see below)

    The configuration is stored in a JSON file at ~/.config/labelr/config.json.

    The following environment variables are supported:
    - LABELR_LABEL_STUDIO_URL
    - LABELR_LABEL_STUDIO_API_KEY
    """
    if CONFIG_PATH.exists():
        config = LabelrConfig.model_validate_json(CONFIG_PATH.read_bytes())

        if "LABELR_LABEL_STUDIO_URL" in os.environ:
            config.label_studio_url = os.environ["LABELR_LABEL_STUDIO_URL"]
        if "LABELR_LABEL_STUDIO_API_KEY" in os.environ:
            config.label_studio_api_key = os.environ["LABELR_LABEL_STUDIO_API_KEY"]
        if "LABELR_LABEL_STUDIO_PROJECT_ID" in os.environ:
            config.label_studio_project_id = int(
                os.environ["LABELR_LABEL_STUDIO_PROJECT_ID"]
            )
        return config
    else:
        return LabelrConfig()


def set_file_config(key: str, value: str):
    """Update the labelr configuration.

    The configuration is stored in a JSON file at ~/.config/labelr/config.json.
    """
    import orjson

    config_json = orjson.loads(CONFIG_PATH.read_bytes())
    config_json[key] = value
    config = LabelrConfig.model_validate(config_json)
    CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
    CONFIG_PATH.write_text(config.model_dump_json(indent=2))


config = get_config()



from typing import Annotated

import typer
from openfoodfacts.utils import get_logger

from labelr.apps import datasets as dataset_app
from labelr.apps import directus as directus_app
from labelr.apps import evaluate as evaluate_app
from labelr.apps import google_batch as google_batch_app
from labelr.apps import hugging_face as hf_app
from labelr.apps import label_studio as ls_app
from labelr.apps import train as train_app
from labelr import config as _config

app = typer.Typer(pretty_exceptions_show_locals=False, no_args_is_help=True)

logger = get_logger()


@app.command()
def predict(
    model_name: Annotated[
        str, typer.Option(help="Name of the object detection model to run")
    ],
    label_names: Annotated[list[str], typer.Argument(help="List of label names")],
    image_url: Annotated[str, typer.Option(help="URL of the image to process")],
    triton_uri: Annotated[
        str, typer.Option(help="URI (host+port) of the Triton Inference Server")
    ],
    image_size: Annotated[
        int, typer.Option(help="Size of the image the model expects")
    ] = 640,
    threshold: Annotated[float, typer.Option(help="Detection threshold")] = 0.5,
    triton_model_version: str = "1",
):
    """Predict objects in an image using an object detection model served by
    Triton."""
    import typing

    from openfoodfacts.ml.object_detection import ObjectDetector
    from openfoodfacts.utils import get_image_from_url
    from PIL import Image

    model = ObjectDetector(
        model_name=model_name, label_names=label_names, image_size=image_size
    )
    image = typing.cast(Image.Image | None, get_image_from_url(image_url))

    if image is None:
        logger.error("Failed to download image from URL: %s", image_url)
        raise typer.Abort()

    output = model.detect_from_image(
        image,
        triton_uri=triton_uri,
        model_version=triton_model_version,
        threshold=threshold,
    )
    results = output.to_list()

    for result in results:
        typer.echo(result)


@app.command()
def config(name: str, value: str):
    """Set a Labelr configuration value.

    Settings can always be overriden using environment variables and command
    option values. The configuration is stored in a JSON file at
    ~/.config/labelr/config.json.

    The following settings can be set:

    - `label_studio_url`: URL of the Label Studio server.
    - `label_studio_api_key`: API key for Label Studio.
    - `project_id`: the ID of the label studio project.

    """
    typer.echo(f"Set '{name}' to '{value}'")
    _config.set_file_config(name, value)
    typer.echo(f"Configuration saved to {_config.CONFIG_PATH}")


app.add_typer(
    ls_app.app,
    name="ls",
    help="Manage Label Studio projects (create, import data, etc.).",
)
app.add_typer(
    hf_app.app,
    name="hf",
    help="Manage Hugging Face Datasets repositories.",
)
app.add_typer(
    dataset_app.app,
    name="datasets",
    help="Manage datasets (convert, export, check, etc.)",
)
app.add_typer(
    train_app.app,
    name="train",
    help="Train models.",
)
app.add_typer(
    evaluate_app.app,
    name="evaluate",
    help="Visualize and evaluate trained models.",
)
app.add_typer(
    google_batch_app.app,
    name="google-batch",
    help="Generate datasets and launch batch jobs on Google Gemini.",
)
app.add_typer(
    directus_app.app, name="directus", help="Manage directus collections and items."
)

if __name__ == "__main__":
    app()



import asyncio
import mimetypes
from collections.abc import Iterator
from pathlib import Path
from typing import Literal
from urllib.parse import urlparse

import aiofiles
import jsonschema
import orjson
import typer
from gcloud.aio.storage import Storage
from openfoodfacts import Flavor
from openfoodfacts.images import generate_image_url
from tqdm.asyncio import tqdm

from labelr.sample.common import SampleMeta
from labelr.sample.llm import LLMImageExtractionSample
from labelr.utils import download_image_from_gcs

try:
    import google.genai  # noqa: F401
except ImportError:
    raise ImportError(
        "The 'google-genai' package is required to use this module. "
        "Please install labelr with the 'google' extra: "
        "`pip install labelr[google]`"
    )
import aiohttp
from google import genai
from google.cloud import storage
from google.genai.types import CreateBatchJobConfig, HttpOptions
from google.genai.types import JSONSchema as GoogleJSONSchema
from google.genai.types import Schema as GoogleSchema
from openfoodfacts.types import JSONType
from pydantic import BaseModel


class RawBatchSamplePart(BaseModel):
    type: Literal["text", "image"]
    data: str


class RawBatchSample(BaseModel):
    key: str
    parts: list[RawBatchSamplePart]
    meta: JSONType = {}


def convert_pydantic_model_to_google_schema(schema: type[BaseModel]) -> JSONType:
    """Google doesn't support natively OpenAPI schemas, so we convert them to
    Google `Schema` (a subset of OpenAPI)."""
    return GoogleSchema.from_json_schema(
        json_schema=GoogleJSONSchema.model_validate(schema.model_json_schema())
    ).model_dump(mode="json", exclude_none=True, exclude_unset=True)


async def download_image(url: str, session: aiohttp.ClientSession) -> bytes:
    """Download an image from a URL and return its content as bytes.

    Args:
        url (str): URL of the image to download.
    Returns:
        bytes: Content of the downloaded image.
    """
    async with session.get(url) as response:
        response.raise_for_status()
        return await response.read()


async def download_image_from_filesystem(url: str, base_dir: Path) -> bytes:
    """Download an image from the filesystem and return its content as bytes.

    Args:
        url (str): URL of the image to download.
        base_dir (Path): Base directory where images are stored.
    Returns:
        bytes: Content of the downloaded image.
    """
    file_path = urlparse(url).path[1:]  # Remove leading '/'
    full_file_path = base_dir / file_path
    async with aiofiles.open(full_file_path, "rb") as f:
        return await f.read()


async def upload_to_gcs(
    image_url: str,
    bucket_name: str,
    blob_name: str,
    session: aiohttp.ClientSession,
    base_image_dir: Path | None = None,
) -> dict:
    """Upload data to Google Cloud Storage.
    Args:
        bucket_name (str): Name of the GCS bucket.
        blob_name (str): Name of the blob (object) in the bucket.
        data (bytes): Data to upload.
        session (aiohttp.ClientSession): HTTP session to use for downloading
            the image.
        base_image_dir (Path | None): If provided, images will be read from
            the filesystem under this base directory instead of downloading
            them from their URLs.
    Returns:
        dict: Status of the upload operation.
    """
    if base_image_dir is None:
        image_data = await download_image(image_url, session)
    else:
        image_data = await download_image_from_filesystem(image_url, base_image_dir)

    client = Storage(session=session)

    status = await client.upload(
        bucket_name,
        blob_name,
        image_data,
    )
    return status


async def upload_to_gcs_format_async(
    sample: RawBatchSample,
    google_json_schema: JSONType,
    instructions: str | None,
    bucket_name: str,
    bucket_dir_name: str,
    session: aiohttp.ClientSession,
    base_image_dir: Path | None = None,
    skip_upload: bool = False,
    thinking_level: str | None = None,
) -> JSONType | None:
    parts: list[JSONType] = []

    if instructions:
        parts.append({"text": instructions})

    for part in sample.parts:
        if part.type == "image":
            mime_type, _ = mimetypes.guess_type(part.data)
            if mime_type is None:
                raise ValueError(f"Cannot guess mimetype of file: {part.data}")

            file_uri = part.data
            image_blob_name = f"{bucket_dir_name}/{sample.key}/{Path(file_uri).name}"
            # Download the image from the URL
            if not skip_upload:
                try:
                    await upload_to_gcs(
                        image_url=file_uri,
                        bucket_name=bucket_name,
                        blob_name=image_blob_name,
                        session=session,
                        base_image_dir=base_image_dir,
                    )
                except FileNotFoundError:
                    return None

            parts.append(
                {
                    "file_data": {
                        "file_uri": f"gs://{bucket_name}/{image_blob_name}",
                        "mime_type": mime_type,
                    }
                }
            )
        else:
            parts.append({"text": part.data})

    generation_config = {
        "responseMimeType": "application/json",
        "response_json_schema": google_json_schema,
    }

    if thinking_level is not None:
        generation_config["thinkingConfig"] = {"thinkingLevel": thinking_level}

    return {
        "key": f"key:{sample.key}",
        "request": {
            "contents": [
                {
                    "parts": parts,
                    "role": "user",
                }
            ],
            "generationConfig": generation_config,
        },
    }


async def generate_batch_dataset(
    data_path: Path,
    output_path: Path,
    google_json_schema: JSONType,
    instructions: str | None,
    bucket_name: str,
    bucket_dir_name: str,
    max_concurrent_uploads: int = 30,
    base_image_dir: Path | None = None,
    from_key: str | None = None,
    skip_upload: bool = False,
    thinking_level: str | None = None,
):
    limiter = asyncio.Semaphore(max_concurrent_uploads)
    ignore = True if from_key is None else False
    missing_files = 0
    async with aiohttp.ClientSession() as session:
        async with asyncio.TaskGroup() as tg:
            async with (
                aiofiles.open(data_path, "r") as input_file,
                aiofiles.open(output_path, "wb") as output_file,
            ):
                async with limiter:
                    tasks = set()
                    async for line in tqdm(input_file, desc="samples"):
                        # print(f"line: {line}")
                        sample = RawBatchSample.model_validate_json(line)
                        # print(f"sample: {sample}")
                        record_key = sample.key
                        if from_key is not None and ignore:
                            if record_key == from_key:
                                ignore = False
                            else:
                                continue
                        task = tg.create_task(
                            upload_to_gcs_format_async(
                                sample=sample,
                                google_json_schema=google_json_schema,
                                instructions=instructions,
                                bucket_name=bucket_name,
                                bucket_dir_name=bucket_dir_name,
                                session=session,
                                base_image_dir=base_image_dir,
                                skip_upload=skip_upload,
                                thinking_level=thinking_level,
                            )
                        )
                        tasks.add(task)

                        if len(tasks) >= max_concurrent_uploads:
                            for task in tasks:
                                await task
                                updated_record = task.result()
                                if updated_record is not None:
                                    await output_file.write(
                                        orjson.dumps(updated_record) + "\n".encode()
                                    )
                                else:
                                    missing_files += 1
                            tasks.clear()

                    for task in tasks:
                        await task
                        updated_record = task.result()
                        if updated_record is not None:
                            await output_file.write(
                                orjson.dumps(updated_record) + "\n".encode()
                            )
                        else:
                            missing_files += 1

    typer.echo(
        f"Upload and dataset update completed. Wrote updated dataset to {output_path}. "
        f"Missing files: {missing_files}."
    )


def launch_batch_job(
    run_name: str,
    dataset_path: Path,
    model: str,
    location: str,
):
    """Launch a Gemini Batch Inference job.

    Args:
        run_name (str): Name of the batch run.
        dataset_path (Path): Path to the dataset file in JSONL format.
        model (str): Model to use for the batch job. Example:
            'gemini-2.5-flash'.
        location (str): Location for the Vertex AI resources. Example:
            'europe-west4'.
    """
    # We upload the dataset to a GCS bucket using the Gcloud

    if model == "gemini-3-pro-preview" and location != "global":
        typer.echo(
            "Warning: only 'global' location is supported for 'gemini-3-pro-preview' model. Overriding location to 'global'."
        )
        location = "global"

    storage_client = storage.Client()
    bucket_name = "robotoff-batch"  # Replace with your bucket name
    run_dir = f"gemini-batch/{run_name}"
    input_file_blob_name = f"{run_dir}/inputs.jsonl"
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(input_file_blob_name)
    blob.upload_from_filename(dataset_path)

    client = genai.Client(
        http_options=HttpOptions(api_version="v1"),
        vertexai=True,
        location=location,
    )
    output_uri = f"gs://{bucket_name}/{run_dir}"
    job = client.batches.create(
        model=model,
        src=f"gs://{bucket_name}/{input_file_blob_name}",
        config=CreateBatchJobConfig(dest=output_uri),
    )
    print(job)


def generate_sample_iter(
    prediction_path: Path,
    json_schema: JSONType,
    skip: int = 0,
    limit: int | None = None,
    is_openfoodfacts_dataset: bool = False,
    openfoodfacts_flavor: Flavor = Flavor.off,
    raise_on_invalid_sample: bool = False,
) -> Iterator[LLMImageExtractionSample]:
    """Generate training samples from a Gemini Batch Inference prediction
    JSONL file.

    Args:
        prediction_path (Path): Path to the prediction JSONL file.
        json_schema (JSONType): JSON schema to validate the predictions.
        skip (int): Number of initial samples to skip.
        limit (int | None): Maximum number of samples to generate.
        is_openfoodfacts_dataset (bool): Whether the dataset is from Open Food
            Facts.
        openfoodfacts_flavor (Flavor): Flavor of the Open Food Facts dataset.
    Yields:
        Iterator[LLMImageExtractionSample]: Generated samples.
    """
    skipped = 0
    invalid = 0
    storage_client = storage.Client()
    with prediction_path.open("r") as f_in:
        for i, sample_str in enumerate(f_in):
            if i < skip:
                skipped += 1
                continue
            if limit is not None and i >= skip + limit:
                break
            sample = orjson.loads(sample_str)
            try:
                yield generate_sample_from_prediction(
                    json_schema=json_schema,
                    sample=sample,
                    is_openfoodfacts_dataset=is_openfoodfacts_dataset,
                    openfoodfacts_flavor=openfoodfacts_flavor,
                    storage_client=storage_client,
                )
            except Exception as e:
                if raise_on_invalid_sample:
                    raise
                else:
                    typer.echo(
                        f"Skipping invalid sample at line {i + 1} in {prediction_path}: {e}"
                    )
                    invalid += 1
                    continue
    if skipped > 0:
        typer.echo(f"Skipped {skipped} samples.")
    if invalid > 0:
        typer.echo(f"Skipped {invalid} invalid samples.")


def generate_sample_from_prediction(
    json_schema: JSONType,
    sample: JSONType,
    is_openfoodfacts_dataset: bool = False,
    openfoodfacts_flavor: Flavor = Flavor.off,
    storage_client: storage.Client | None = None,
) -> LLMImageExtractionSample:
    """Generate a LLMImageExtractionSample from a prediction sample.
    Args:
        json_schema (JSONType): JSON schema to validate the predictions.
        sample (JSONType): Prediction sample.
        is_openfoodfacts_dataset (bool): Whether the dataset is from Open Food
            Facts.
        openfoodfacts_flavor (Flavor): Flavor of the Open Food Facts dataset.
        storage_client (storage.Client | None): Optional Google Cloud Storage
            client. If not provided, a new client will be created.
    Returns:
        LLMImageExtractionSample: Generated sample.
    """
    image_id = sample["key"][len("key:") :]
    response_str = sample["response"]["candidates"][0]["content"]["parts"][0]["text"]
    image_uri = sample["request"]["contents"][0]["parts"][1]["file_data"]["file_uri"]
    image = download_image_from_gcs(image_uri=image_uri, client=storage_client)
    response = orjson.loads(response_str)
    jsonschema.validate(response, json_schema)

    if is_openfoodfacts_dataset:
        image_stem_parts = image_id.split("_")
        barcode = image_stem_parts[0]
        off_image_id = image_stem_parts[1]
        image_id = f"{barcode}_{off_image_id}"
        image_url = generate_image_url(
            barcode, off_image_id, flavor=openfoodfacts_flavor
        )
    else:
        image_id = image_id
        barcode = ""
        off_image_id = ""
        image_url = ""

    sample_meta = SampleMeta(
        barcode=barcode,
        off_image_id=off_image_id,
        image_url=image_url,
    )
    return LLMImageExtractionSample(
        image_id=image_id,
        image=image,
        output=orjson.dumps(response).decode("utf-8"),
        meta=sample_meta,
    )



import json
from pathlib import Path
from typing import Annotated

import typer
from openfoodfacts.utils import get_logger

from . import typer_description
from ...config import config, check_required_field
from .common import check_label_studio_api_key
from .prediction import app as prediction_app
from .tag import app as tag_app

app = typer.Typer(no_args_is_help=True)
app.add_typer(
    prediction_app, name="prediction", help="Manage predictions on Label Studio."
)
app.add_typer(tag_app, name="tag", help="Manage tags on tasks.")

logger = get_logger(__name__)


@app.command()
def create(
    title: Annotated[str, typer.Option(help="Project title")],
    config_file: Annotated[
        Path, typer.Option(help="Path to label config file", file_okay=True)
    ],
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
):
    """Create a new Label Studio project."""
    from label_studio_sdk.client import LabelStudio

    check_label_studio_api_key(api_key)
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)
    label_config = config_file.read_text()

    project = ls.projects.create(title=title, label_config=label_config)
    logger.info(f"Project created: {project}")


@app.command()
def import_data(
    dataset_path: Annotated[
        Path,
        typer.Option(
            help="Path to the Label Studio dataset JSONL file", file_okay=True
        ),
    ],
    project_id: Annotated[
        int | None, typer.Option(help=typer_description.LABEL_STUDIO_PROJECT_ID)
    ] = config.label_studio_project_id,
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
    batch_size: Annotated[
        int, typer.Option(help="Number of tasks to import as a single batch")
    ] = 25,
):
    """Import tasks from a dataset file to a Label Studio project.

    The dataset file must be a JSONL file: it should contain one JSON object
    per line. To generate such a file, you can use the `create-dataset-file`
    command.
    """
    import more_itertools
    import tqdm
    from label_studio_sdk.client import LabelStudio

    check_label_studio_api_key(api_key)
    check_required_field("--project-id", project_id)
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)

    with dataset_path.open("rt") as f:
        for batch in more_itertools.chunked(
            tqdm.tqdm(map(json.loads, f), desc="tasks"), batch_size
        ):
            ls.projects.import_tasks(id=project_id, request=batch)


@app.command()
def add_split(
    train_split: Annotated[
        float, typer.Option(help="fraction of samples to add in train split")
    ],
    project_id: Annotated[
        int | None, typer.Option(help=typer_description.LABEL_STUDIO_PROJECT_ID)
    ] = config.label_studio_project_id,
    train_split_name: Annotated[
        str,
        typer.Option(help="name of the train split"),
    ] = "train",
    val_split_name: Annotated[
        str,
        typer.Option(help="name of the validation split"),
    ] = "val",
    task_id_file: Annotated[
        Path | None,
        typer.Option(help="path of a text file containing IDs of samples"),
    ] = None,
    split_name: Annotated[
        str | None,
        typer.Option(
            help="name of the split associated "
            "with the task ID file. If --task-id-file is not provided, "
            "this field is ignored."
        ),
    ] = None,
    overwrite: Annotated[
        bool, typer.Option(help="overwrite existing split field")
    ] = False,
    view_id: Annotated[
        int | None,
        typer.Option(
            help="ID of the Label Studio view, if any. This option is useful "
            "to filter the task to process."
        ),
    ] = None,
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
):
    """Update the split field of tasks in a Label Studio project.

    The behavior of this command depends on the `--task-id-file` option.

    If `--task-id-file` is provided, it should contain a list of task IDs,
    one per line. The split field of these tasks will be updated to the value
    of `--split-name`. The `--train-split` value is ignored in this case.

    If `--task-id-file` is not provided, the split field of all tasks in the
    project will be updated based on the `train_split` probability.
    The split field is set to "train" with probability `train_split`, and "val"
    otherwise.

    In both cases, tasks with a non-null split field are not updated unless
    the `--overwrite` flag is provided.

    The `--view-id` option can be used to only assign the split on a subset
    of the tasks.
    """
    import random

    from label_studio_sdk.types import LseTask
    from label_studio_sdk.client import LabelStudio

    check_label_studio_api_key(api_key)
    check_required_field("--project-id", project_id)
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)

    task_ids = None
    if task_id_file is not None:
        if split_name is None or split_name not in (train_split_name, val_split_name):
            raise typer.BadParameter(
                "--split-name is required when using --task-id-file"
            )
        task_ids = task_id_file.read_text().strip().split("\n")

    for task in ls.tasks.list(project=project_id, fields="all", view=view_id):
        task: LseTask
        task_id = task.id

        split = task.data.get("split")
        if split is None or overwrite:
            if task_ids:
                if str(task_id) in task_ids:
                    split = split_name
                else:
                    continue
            else:
                split = (
                    train_split_name
                    if random.random() < train_split
                    else val_split_name
                )

            logger.info("Updating task: %s, split: %s", task.id, split)
            ls.tasks.update(task.id, data={**task.data, "split": split})


@app.command()
def create_dataset_file(
    input_file: Annotated[
        Path,
        typer.Option(
            help="Path to the input file. Can be a text or JSONL file.",
            exists=True,
            readable=True,
            dir_okay=False,
        ),
    ],
    output_file: Annotated[
        Path,
        typer.Option(help="Path to the output JSONL file", exists=False, writable=True),
    ],
):
    """Create a Label Studio object detection dataset file from file, containing a list
    a list of image URLs.

    The file can be a plain text file, in which case we expect a single image URL
    per line. Labelr set the `image_id` field in Label Studio to the URL filename
    (stripped from its extension).

    The file can also be a JSONL file, with each line of the file being a JSON
    containing:

    - `image_url` (str, required): the URL of the image
    - `image_id` (str, optional): the ID of the image, used as value for the
        `image_id` in Label Studio. If not provided, we use the stemmed name of the
        file in the URL as `image_id`.

    Other fields in the JSON will be added to the `meta` field of each task.

    The output file is a JSONL file. It cannot be imported directly in Label
    Studio (which requires a JSON file as input), the `import-data` command
    should be used to import the generated dataset file.
    """
    from urllib.parse import urlparse
    import typing

    from PIL import Image
    import tqdm
    from openfoodfacts.images import extract_barcode_from_url, extract_source_from_url
    from openfoodfacts.utils import get_image_from_url
    from openfoodfacts.types import JSONType
    import orjson

    from labelr.sample.object_detection import format_object_detection_sample_to_ls

    logger.info("Loading dataset: %s", input_file)

    jsonl_mode = output_file.name.endswith(".jsonl")

    with output_file.open("wt") as f:
        for line in tqdm.tqdm(input_file.open("rt"), desc="images"):
            extra_meta: JSONType = {}
            if jsonl_mode:
                item = orjson.loads(line)
                url = item.pop("image_url")
                image_id = item.pop("image_id", None)
                extra_meta |= item
            else:
                url = line.strip()
                image_id = None
                if not url:
                    continue

            if not image_id:
                image_id = Path(urlparse(url).path).stem

            if ".openfoodfacts.org" in url:
                barcode = extract_barcode_from_url(url)
                extra_meta["barcode"] = barcode
                off_image_id = Path(extract_source_from_url(url)).stem
                extra_meta["off_image_id"] = off_image_id
                image_id = f"{barcode}_{off_image_id}"

            image = typing.cast(Image.Image, get_image_from_url(url, error_raise=False))

            if image is None:
                logger.warning("Failed to load image: %s", url)
                continue

            label_studio_sample = format_object_detection_sample_to_ls(
                image_id, url, image.width, image.height, extra_meta
            )
            f.write(json.dumps(label_studio_sample) + "\n")


@app.command()
def create_config_file(
    output_file: Annotated[
        Path, typer.Option(help="Path to the output label config file", exists=False)
    ],
    labels: Annotated[
        list[str], typer.Option(help="List of class labels to use for the model")
    ],
):
    """Create a Label Studio label config file for object detection tasks."""
    from labelr.project_config import create_object_detection_label_config

    config = create_object_detection_label_config(labels)
    output_file.write_text(config)
    logger.info("Label config file created: %s", output_file)


@app.command()
def check_dataset(
    project_id: Annotated[
        int | None, typer.Option(help=typer_description.LABEL_STUDIO_PROJECT_ID)
    ] = config.label_studio_project_id,
    view_id: Annotated[int, typer.Option(help="Label Studio View ID, if any.")] = None,
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
    delete_missing_images: Annotated[
        bool,
        typer.Option(help="Delete tasks with missing images from the dataset"),
    ] = False,
    delete_duplicate_images: Annotated[
        bool, typer.Option(help="Delete duplicate images from the dataset")
    ] = False,
):
    """Perform sanity checks of a Label Studio dataset.

    This function checks for:
    - Tasks with missing images (404)
    - Duplicate images based on perceptual hash (pHash)
    - Tasks with multiple annotations

    This function doesn't perform any modifications to the dataset, except
    optionally deleting tasks with missing images if --delete-missing-images
    is provided and tasks with duplicate images if --delete-duplicate-images
    is provided.
    """
    import typing
    from label_studio_sdk.client import LabelStudio

    from labelr.check import check_ls_dataset

    check_label_studio_api_key(api_key)
    check_required_field("--project-id", project_id)
    project_id = typing.cast(int, project_id)
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)
    check_ls_dataset(
        ls=ls,
        project_id=project_id,
        view_id=view_id,
        delete_missing_images=delete_missing_images,
        delete_duplicate_images=delete_duplicate_images,
    )


@app.command()
def list_users(
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
):
    """List all users in Label Studio."""
    from label_studio_sdk.client import LabelStudio

    check_label_studio_api_key(api_key)
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)

    for user in ls.users.list():
        print(f"{user.id:02d}: {user.email}")


@app.command()
def delete_user(
    user_id: int,
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
):
    """Delete a user from Label Studio."""
    from label_studio_sdk.client import LabelStudio

    check_label_studio_api_key(api_key)
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)
    ls.users.delete(user_id)


@app.command()
def dump_dataset(
    output_file: Annotated[
        Path, typer.Option(help="Path of the output file", writable=True)
    ],
    project_id: Annotated[
        int | None, typer.Option(help=typer_description.LABEL_STUDIO_PROJECT_ID)
    ] = config.label_studio_project_id,
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    view_id: Annotated[
        int | None,
        typer.Option(
            help="ID of the Label Studio view, if any. This option is useful "
            "to filter the tasks to dump."
        ),
    ] = None,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
):
    """Dump all the tasks of a dataset in a JSONL file.

    All fields of the tasks are exported. A subset of the tasks can be
    selected by filtering tasks based on a view (=tab) using the `--view-id`
    option.
    """
    import orjson
    import tqdm
    from label_studio_sdk.client import LabelStudio

    check_label_studio_api_key(api_key)
    check_required_field("--project-id", project_id)
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)

    with output_file.open("wb") as f:
        for task in tqdm.tqdm(
            ls.tasks.list(project=project_id, view=view_id), desc="tasks"
        ):
            content = orjson.dumps(task.dict())
            f.write(content + b"\n")



import datasets
from datasets import Features
from datasets import Image as HFImage

# dataset features for predictions.parquet used in evaluation and visualization
OBJECT_DETECTION_DS_PREDICTION_FEATURES = Features(
    {
        "image": HFImage(),
        "image_with_prediction": HFImage(),
        "image_id": datasets.Value("string"),
        "detected": {
            "bbox": datasets.Sequence(datasets.Sequence(datasets.Value("float32"))),
            "category_id": datasets.Sequence(datasets.Value("int64")),
            "category_name": datasets.Sequence(datasets.Value("string")),
            "confidence": datasets.Sequence(datasets.Value("float32")),
        },
        "split": datasets.Value("string"),
        "width": datasets.Value("int64"),
        "height": datasets.Value("int64"),
        "meta": {
            "barcode": datasets.Value("string"),
            "off_image_id": datasets.Value("string"),
            "image_url": datasets.Value("string"),
        },
        "objects": {
            "bbox": datasets.Sequence(datasets.Sequence(datasets.Value("float32"))),
            "category_id": datasets.Sequence(datasets.Value("int64")),
            "category_name": datasets.Sequence(datasets.Value("string")),
        },
    }
)



from typing import Annotated

import typer

app = typer.Typer(no_args_is_help=True)


@app.command()
def visualize_object_detection(
    hf_repo_id: Annotated[
        str,
        typer.Option(
            ...,
            help="Hugging Face repository ID of the trained model. "
            "A `predictions.parquet` file is expected in the repo. Revision can be specified "
            "by appending `@<revision>` to the repo ID.",
        ),
    ],
    dataset_name: Annotated[
        str | None, typer.Option(..., help="Name of the FiftyOne dataset to create.")
    ] = None,
    persistent: Annotated[
        bool,
        typer.Option(
            ...,
            help="Whether to make the FiftyOne dataset persistent (i.e., saved to disk).",
        ),
    ] = False,
):
    """Visualize object detection model predictions stored in a Hugging Face
    repository using FiftyOne."""
    from labelr.evaluate import object_detection

    if dataset_name is None:
        dataset_name = hf_repo_id.replace("/", "-").replace("@", "-")

    object_detection.visualize(
        hf_repo_id=hf_repo_id,
        dataset_name=dataset_name,
        persistent=persistent,
    )



"""Commands to manage datasets local datasets and export between platforms
(Label Studio, HuggingFace Hub, local dataset,...)."""

import json
import random
import shutil
import typing
from pathlib import Path
from typing import Annotated, Optional

import typer
from openfoodfacts import Flavor
from openfoodfacts.utils import get_logger

from .label_studio import typer_description
from ..config import config, check_required_field
from ..types import ExportDestination, ExportSource, TaskType

app = typer.Typer(no_args_is_help=True)

logger = get_logger(__name__)


@app.command()
def check(
    dataset_dir: Annotated[
        Path,
        typer.Option(
            help="Path to the dataset directory", exists=True, file_okay=False
        ),
    ],
    remove: Annotated[
        bool,
        typer.Option(help="Remove duplicate images from the dataset"),
    ] = False,
):
    """Check a local dataset in Ultralytics format for duplicate images."""

    from ..check import check_local_dataset

    check_local_dataset(dataset_dir, remove=remove)


@app.command()
def split_train_test(
    task_type: TaskType, dataset_dir: Path, output_dir: Path, train_ratio: float = 0.8
):
    """Split a local dataset into training and test sets.

    Only classification tasks are supported.
    """
    if task_type == TaskType.classification:
        class_dirs = [d for d in dataset_dir.iterdir() if d.is_dir()]
        logger.info("Found classes: %s", [d.name for d in class_dirs])

        output_dir.mkdir(parents=True, exist_ok=True)
        train_dir = output_dir / "train"
        test_dir = output_dir / "test"
        train_dir.mkdir(parents=True, exist_ok=True)
        test_dir.mkdir(parents=True, exist_ok=True)

        for class_dir in class_dirs:
            input_paths = list(class_dir.glob("*"))
            random.shuffle(input_paths)

            test_count = int(len(input_paths) * (1 - train_ratio))
            if test_count == 0:
                logger.warning("Not enough samples, skipping class: %s", class_dir.name)
                continue

            test_paths = input_paths[:test_count]
            train_paths = input_paths[test_count:]

            for output_dir, input_paths in (
                (train_dir, train_paths),
                (test_dir, test_paths),
            ):
                output_cls_dir = output_dir / class_dir.name
                output_cls_dir.mkdir(parents=True, exist_ok=True)

                for path in input_paths:
                    logger.info("Copying: %s to %s", path, output_cls_dir)
                    shutil.copy(path, output_cls_dir / path.name)
    else:
        raise typer.BadParameter("Unsupported task type")


@app.command()
def convert_object_detection_dataset(
    repo_id: Annotated[
        str, typer.Option(help="Hugging Face Datasets repository ID to convert")
    ],
    output_file: Annotated[
        Path, typer.Option(help="Path to the output JSON file", exists=False)
    ],
):
    """Convert object detection dataset from Hugging Face Datasets to Label
    Studio format, and save it to a JSON file."""
    from datasets import load_dataset

    from labelr.sample.object_detection import (
        format_object_detection_sample_from_hf_to_ls,
    )

    logger.info("Loading dataset: %s", repo_id)
    ds = load_dataset(repo_id)
    logger.info("Dataset loaded: %s", tuple(ds.keys()))

    with output_file.open("wt") as f:
        for split in ds.keys():
            logger.info("Processing split: %s", split)
            for sample in ds[split]:
                label_studio_sample = format_object_detection_sample_from_hf_to_ls(
                    sample, split=split
                )
                f.write(json.dumps(label_studio_sample) + "\n")


@app.command()
def export(
    from_: Annotated[ExportSource, typer.Option("--from", help="Input source to use.")],
    to: Annotated[ExportDestination, typer.Option(help="Where to export the data.")],
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    task_type: Annotated[
        TaskType, typer.Option(help="Type of task to export.")
    ] = TaskType.object_detection,
    repo_id: Annotated[
        str | None,
        typer.Option(
            help="Hugging Face Datasets repository ID to convert (only if --from or --to is `hf`)."
        ),
    ] = None,
    label_names: Annotated[
        str | None,
        typer.Option(help="Label names to use, as a comma-separated list"),
    ] = None,
    skip_labels: Annotated[
        str | None,
        typer.Option(
            help="Label names to skip  as a comma-separated list. If the label of an "
            "annotation result is in this list, it will be skipped. Only used if the "
            "source is Label Studio and the task type is object detection."
        ),
    ] = None,
    project_id: Annotated[
        int | None, typer.Option(help=typer_description.LABEL_STUDIO_PROJECT_ID)
    ] = config.label_studio_project_id,
    view_id: Annotated[
        int | None,
        typer.Option(
            help="ID of the Label Studio view, if any. This option is useful "
            "to filter the task to export."
        ),
    ] = None,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
    output_dir: Annotated[
        Path | None,
        typer.Option(
            help="Path to the output directory. Only used if the destintation (`to`) is `ultralytics`",
            file_okay=False,
        ),
    ] = None,
    dataset_dir: Annotated[
        Path | None,
        typer.Option(help="Path to the dataset directory, only for Ultralytics source"),
    ] = None,
    download_images: Annotated[
        bool,
        typer.Option(
            help="if True, don't use HF images and download images from the server"
        ),
    ] = False,
    is_openfoodfacts_dataset: Annotated[
        bool,
        typer.Option(
            help="Whether the Ultralytics dataset is an Open Food Facts dataset, only "
            "for Ultralytics source. This is used:\n"
            "- to generate the correct image URLs from each image name, when exporting "
            "from Ultralytics to Hugging Face Datasets.\n"
            "- to include additional metadata fields specific to Open Food Facts "
            "(`barcode` and `off_image_id`) when exporting from Label Studio to "
            "Hugging Face Datasets."
        ),
    ] = False,
    openfoodfacts_flavor: Annotated[
        Flavor,
        typer.Option(
            help="Flavor of the Open Food Facts dataset to use for image URLs, only "
            "for Ultralytics source if is_openfoodfacts_dataset is True. This is used to "
            "generate the correct image URLs each image name. This option is ignored if "
            "is_openfoodfacts_dataset is False."
        ),
    ] = Flavor.off,
    train_ratio: Annotated[
        float,
        typer.Option(
            help="Train ratio for splitting the dataset, if the split name is not "
            "provided. Only used if the source is Label Studio and the destination "
            "is Ultralytics."
        ),
    ] = 0.8,
    image_max_size: Annotated[
        int | None,
        typer.Option(
            help="Maximum size (in pixels) for the images. If None, no resizing is performed."
            "Otherwise, the longest side of the image will be resized to this value, "
            "keeping the aspect ratio."
        ),
    ] = None,
    error_raise: Annotated[
        bool,
        typer.Option(
            help="Raise an error if an image download fails, only for Ultralytics"
        ),
    ] = True,
    use_aws_cache: Annotated[
        bool,
        typer.Option(
            help="Use the AWS S3 cache for image downloads instead of images.openfoodfacts.org, "
            "it is ignored if the export format is not Ultralytics"
        ),
    ] = False,
    merge_labels: Annotated[
        bool,
        typer.Option(help="Merge multiple labels into a single label"),
    ] = False,
    revision: Annotated[
        str,
        typer.Option(
            help="Revision (branch, tag or commit) for the Hugging Face Datasets repository. "
            "Only used when importing from or exporting to Hugging Face Datasets."
        ),
    ] = "main",
):
    """Export Label Studio annotation, either to Hugging Face Datasets or
    local files (ultralytics format)."""
    from label_studio_sdk.client import LabelStudio

    from labelr.export.object_detection import (
        export_from_hf_to_ultralytics_object_detection,
    )
    from labelr.export.object_detection import (
        export_from_ls_to_hf_object_detection,
        export_from_ls_to_ultralytics_object_detection,
    )
    from labelr.export.common import export_from_ultralytics_to_hf

    check_required_field("--project-id", project_id)

    if (to == ExportDestination.hf or from_ == ExportSource.hf) and repo_id is None:
        raise typer.BadParameter("Repository ID is required for export/import with HF")

    if from_ == ExportSource.ultralytics and dataset_dir is None:
        raise typer.BadParameter(
            "Dataset directory is required for export from Ultralytics source"
        )

    label_names_list: list[str] | None = None

    if label_names is None:
        if to == ExportDestination.hf:
            raise typer.BadParameter("Label names are required for HF export")
        if from_ == ExportSource.ls:
            raise typer.BadParameter(
                "Label names are required for export from LS source"
            )
    else:
        label_names = typing.cast(str, label_names)
        label_names_list = label_names.split(",")

    skip_labels_list = []
    if skip_labels is not None:
        skip_labels = typing.cast(str, skip_labels)
        skip_labels_list = skip_labels.split(",")

    if from_ == ExportSource.ls:
        if project_id is None:
            raise typer.BadParameter("Project ID is required for LS export")
        if api_key is None:
            raise typer.BadParameter("API key is required for LS export")

    if to == ExportDestination.ultralytics and output_dir is None:
        raise typer.BadParameter("Output directory is required for Ultralytics export")

    if from_ == ExportSource.ls:
        if task_type != TaskType.object_detection:
            raise typer.BadParameter(
                "Only object detection task is currently supported with LS source"
            )
        ls = LabelStudio(base_url=label_studio_url, api_key=api_key)
        if to == ExportDestination.hf:
            repo_id = typing.cast(str, repo_id)
            export_from_ls_to_hf_object_detection(
                ls,
                repo_id=repo_id,
                label_names=typing.cast(list[str], label_names_list),
                project_id=typing.cast(int, project_id),
                is_openfoodfacts_dataset=is_openfoodfacts_dataset,
                merge_labels=merge_labels,
                use_aws_cache=use_aws_cache,
                revision=revision,
                view_id=view_id,
                image_max_size=image_max_size,
                skip_labels=skip_labels_list,
            )
        elif to == ExportDestination.ultralytics:
            export_from_ls_to_ultralytics_object_detection(
                ls,
                typing.cast(Path, output_dir),
                typing.cast(list[str], label_names_list),
                typing.cast(int, project_id),
                train_ratio=train_ratio,
                error_raise=error_raise,
                merge_labels=merge_labels,
                use_aws_cache=use_aws_cache,
                view_id=view_id,
                image_max_size=image_max_size,
                skip_labels=skip_labels_list,
            )

    elif from_ == ExportSource.hf:
        if task_type != TaskType.object_detection:
            raise typer.BadParameter(
                "Only object detection task is currently supported with HF source"
            )
        if to == ExportDestination.ultralytics:
            export_from_hf_to_ultralytics_object_detection(
                typing.cast(str, repo_id),
                typing.cast(Path, output_dir),
                download_images=download_images,
                error_raise=error_raise,
                use_aws_cache=use_aws_cache,
                revision=revision,
                image_max_size=image_max_size,
                skip_labels=skip_labels_list,
            )
        else:
            raise typer.BadParameter("Unsupported export format")
    elif from_ == ExportSource.ultralytics:
        if task_type != TaskType.classification:
            raise typer.BadParameter(
                "Only classification task is currently supported with Ultralytics source"
            )
        if to == ExportDestination.hf:
            export_from_ultralytics_to_hf(
                task_type=task_type,
                dataset_dir=typing.cast(Path, dataset_dir),
                repo_id=typing.cast(str, repo_id),
                merge_labels=merge_labels,
                label_names=typing.cast(list[str], label_names_list),
                is_openfoodfacts_dataset=is_openfoodfacts_dataset,
                openfoodfacts_flavor=openfoodfacts_flavor,
            )


@app.command()
def export_llm_ds(
    dataset_path: Annotated[
        Path, typer.Option(..., help="Path to the JSONL dataset file")
    ],
    repo_id: Annotated[
        str, typer.Option(..., help="Hugging Face Datasets repository ID to export to")
    ],
    split: Annotated[str, typer.Option(..., help="Dataset split to export")],
    revision: Annotated[
        str,
        typer.Option(
            help="Revision (branch, tag or commit) for the Hugging Face Datasets repository."
        ),
    ] = "main",
    tmp_dir: Annotated[
        Path | None,
        typer.Option(
            help="Path to the temporary directory used to store intermediate sample files "
            "created when building the HF dataset.",
        ),
    ] = None,
    image_max_size: Annotated[
        int | None,
        typer.Option(
            help="Maximum size (in pixels) for the images. If None, no resizing is performed.",
        ),
    ] = None,
):
    """Export LLM image extraction dataset with images only to Hugging Face
    Datasets.
    """
    from labelr.export.llm import export_to_hf_llm_image_extraction
    from labelr.sample.llm import load_llm_image_extraction_dataset_from_jsonl

    sample_iter = load_llm_image_extraction_dataset_from_jsonl(
        dataset_path=dataset_path
    )
    export_to_hf_llm_image_extraction(
        sample_iter,
        split=split,
        repo_id=repo_id,
        revision=revision,
        tmp_dir=tmp_dir,
        image_max_size=image_max_size,
    )


@app.command()
def update_llm_ds(
    dataset_path: Annotated[
        Path, typer.Option(help="Path to the JSONL containing the updates.")
    ],
    repo_id: Annotated[
        str, typer.Option(help="Hugging Face Datasets repository ID to update")
    ],
    split: Annotated[str, typer.Option(help="Dataset split to use")],
    revision: Annotated[
        str,
        typer.Option(
            help="Revision (branch, tag or commit) to use when pushing the new version "
            "of the Hugging Face Dataset."
        ),
    ] = "main",
    tmp_dir: Annotated[
        Path | None,
        typer.Option(
            help="Path to a temporary directory to use for image processing",
        ),
    ] = None,
    show_diff: Annotated[
        bool,
        typer.Option(
            help="Show the differences between the original sample and the update. If "
            "True, the updated dataset is not pushed to the Hub. Useful to review the "
            "updates before applying them.",
        ),
    ] = False,
):
    """Update an existing LLM image extraction dataset, by updating the
    `output` field of each sample in the dataset.

    The `--dataset_path` JSONL file should contain items with two fields:

    - `image_id`: The image ID of the sample to update in the Hugging Face
        dataset.
    - `output`: The new output data to set for the sample.
    """
    import sys
    from difflib import Differ

    import orjson
    from datasets import load_dataset
    from diskcache import Cache

    dataset = load_dataset(repo_id, split=split)

    # Populate cache with the updates
    cache = Cache(directory=tmp_dir or None)
    with dataset_path.open("r") as f:
        for line in map(orjson.loads, f):
            if "image_id" not in line or "output" not in line:
                raise ValueError(
                    "Each item in the update JSONL file must contain `image_id` and `output` fields"
                )
            image_id = line["image_id"]
            output = line["output"]

            if not isinstance(output, str):
                output = orjson.dumps(output).decode("utf-8")

            cache[image_id] = output

    def apply_updates(sample):
        image_id = sample["image_id"]
        if image_id in cache:
            cached_item = cache[image_id]
            sample["output"] = cached_item
        return sample

    if show_diff:
        differ = Differ()
        for sample in dataset:
            image_id = sample["image_id"]
            if image_id in cache:
                cached_item = orjson.loads(cache[image_id])
                original_item = orjson.loads(sample["output"])
                cached_item_str = orjson.dumps(
                    cached_item, option=orjson.OPT_INDENT_2
                ).decode("utf8")
                original_item_str = orjson.dumps(
                    original_item, option=orjson.OPT_INDENT_2
                ).decode("utf8")
                diff = list(
                    differ.compare(
                        original_item_str.splitlines(keepends=True),
                        cached_item_str.splitlines(keepends=True),
                    )
                )
                sys.stdout.writelines(diff)
                sys.stdout.write("\n" + "-" * 30 + "\n")

    else:
        updated_dataset = dataset.map(apply_updates, batched=False)
        updated_dataset.push_to_hub(repo_id, split=split, revision=revision)



import functools
import logging
import pickle
import tempfile
import typing
from collections.abc import Iterator
from pathlib import Path

import datasets
import tqdm
from PIL import Image, ImageOps

from labelr.export.common import _pickle_sample_generator
from labelr.sample.llm import (
    HF_DS_LLM_IMAGE_EXTRACTION_FEATURES,
    LLMImageExtractionSample,
)
from labelr.utils import PathWithContext

logger = logging.getLogger(__name__)


def export_to_hf_llm_image_extraction(
    sample_iter: Iterator[LLMImageExtractionSample],
    split: str,
    repo_id: str,
    revision: str = "main",
    tmp_dir: Path | None = None,
    image_max_size: int | None = None,
) -> None:
    """Export LLM image extraction samples to a Hugging Face dataset.

    Args:
        sample_iter (Iterator[LLMImageExtractionSample]): Iterator of samples
            to export.
        split (str): Name of the dataset split (e.g., 'train', 'val').
        repo_id (str): Hugging Face repository ID to push the dataset to.
        revision (str): Revision (branch, tag or commit) to use for the
            Hugging Face Datasets repository.
        tmp_dir (Path | None): Temporary directory to use for intermediate
            files. If None, a temporary directory will be created
            automatically.
        image_max_size (int | None): Maximum size (in pixels) for the images.
    """
    logger.info(
        "Repo ID: %s, revision: %s, split: %s, tmp_dir: %s, image_max_size: %s",
        repo_id,
        revision,
        split,
        tmp_dir,
        image_max_size,
    )

    tmp_dir_with_context: PathWithContext | tempfile.TemporaryDirectory
    if tmp_dir:
        tmp_dir.mkdir(parents=True, exist_ok=True)
        tmp_dir_with_context = PathWithContext(tmp_dir)
    else:
        tmp_dir_with_context = tempfile.TemporaryDirectory()

    with tmp_dir_with_context as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        for sample in tqdm.tqdm(sample_iter, desc="samples"):
            image = sample.image
            # Rotate image according to exif orientation using Pillow
            image = typing.cast(Image.Image, ImageOps.exif_transpose(image))

            if image_max_size is not None:
                if image.height > image_max_size or image.width > image_max_size:
                    image.thumbnail(
                        (image_max_size, image_max_size),
                        Image.Resampling.LANCZOS,
                    )
            image_id = sample.image_id
            json_sample = {
                "image_id": image_id,
                "image": image,
                "meta": {
                    k: v for k, v in sample.meta.model_dump().items() if v is not None
                },
                "output": sample.output,
            }
            # Save output as pickle
            with open(tmp_dir / f"{split}_{image_id}.pkl", "wb") as f:
                pickle.dump(json_sample, f)

        hf_ds = datasets.Dataset.from_generator(
            functools.partial(_pickle_sample_generator, tmp_dir),
            features=HF_DS_LLM_IMAGE_EXTRACTION_FEATURES,
        )
        hf_ds.push_to_hub(repo_id, split=split, revision=revision)



import enum


class ExportSource(str, enum.Enum):
    hf = "hf"
    ls = "ls"
    ultralytics = "ultralytics"


class ExportDestination(str, enum.Enum):
    hf = "hf"
    ultralytics = "ultralytics"


class TaskType(str, enum.Enum):
    object_detection = "object_detection"
    classification = "classification"



import asyncio
import importlib
from pathlib import Path
from typing import Annotated, Any

import typer
from openfoodfacts import Flavor
from openfoodfacts.types import JSONType
from pydantic import BaseModel


app = typer.Typer(no_args_is_help=True)


def _check_json_schema(item: JSONType) -> None:
    if item.get("type") == "object":
        required_fields = item.get("required", [])
        all_fields = item.get("properties", [])
        diff = set(all_fields) - set(required_fields)
        if diff:
            raise ValueError(
                f"fields '{diff}' must be marked as required in the JSONSchema. "
                "All fields with type 'object' must be required."
            )
    return None


def check_json_schema(json_schema: JSONType) -> None:
    """Check that for all `object`s, all fields are marked as required.

    This is important to check, as otherwise the structured generation
    backend may prevent the model to generate these fields.
    This is the case as of vLLM 0.13 and xgrammars as backend.

    To prevent this, we ask all fields to be marked as required.
    """
    stack = [json_schema]

    for def_item in json_schema.get("$defs", {}).values():
        stack.append(def_item)

    while stack:
        item = stack.pop()
        _check_json_schema(item)
        for sub_item in item.get("properties", {}).values():
            stack.append(sub_item)


def convert_pydantic_model_to_google_schema(schema: type[BaseModel]) -> dict[str, Any]:
    """Google doesn't support natively OpenAPI schemas, so we convert them to
    Google `Schema` (a subset of OpenAPI)."""
    from google.genai.types import JSONSchema as GoogleJSONSchema
    from google.genai.types import Schema as GoogleSchema

    return GoogleSchema.from_json_schema(
        json_schema=GoogleJSONSchema.model_validate(schema.model_json_schema())
    ).model_dump(mode="json", exclude_none=True, exclude_unset=True)


@app.command()
def generate_dataset(
    data_path: Annotated[
        Path,
        typer.Option(
            ...,
            help="Path to a JSONL file containing the raw batch samples.",
            exists=True,
            dir_okay=False,
            resolve_path=True,
        ),
    ],
    output_path: Annotated[
        Path,
        typer.Option(
            ...,
            help="Path where to write the generated dataset file.",
            exists=False,
            dir_okay=False,
            resolve_path=True,
        ),
    ],
    config_module: Annotated[
        str,
        typer.Option(
            ...,
            help="Python module path (e.g., 'myschema') containing two variables: "
            "OUTPUT_SCHEMA (a Pydantic class representing the output schema) and "
            "INSTRUCTIONS (a str containing instructions to add before each sample).",
        ),
    ],
    bucket_name: Annotated[
        str,
        typer.Option(
            ...,
            help="Name of the GCS bucket where the images are stored.",
        ),
    ] = "robotoff-batch",
    bucket_dir_name: Annotated[
        str,
        typer.Option(
            ...,
            help="Directory name in the GCS bucket where the images are stored.",
        ),
    ] = "gemini-batch-images",
    max_concurrent_uploads: Annotated[
        int,
        typer.Option(
            ...,
            help="Maximum number of concurrent uploads to GCS.",
        ),
    ] = 30,
    base_image_dir: Annotated[
        Path | None,
        typer.Option(
            ...,
            help="Base directory to resolve local image paths from.",
        ),
    ] = None,
    from_key: Annotated[
        str | None,
        typer.Option(
            ...,
            help="If specified, resume processing from this sample key.",
        ),
    ] = None,
    skip_upload: Annotated[
        bool, typer.Option(..., help="Skip uploading images to GCS")
    ] = False,
    thinking_level: Annotated[
        str | None,
        typer.Option(
            ...,
            help="Thinking level to use for the generation config.",
        ),
    ] = None,
):
    """Generate a dataset file in JSONL format to be used for batch
    processing, using Gemini Batch Inference."""
    from labelr.google_genai import generate_batch_dataset

    typer.echo(f"Uploading images from '{data_path}' to GCS bucket '{bucket_name}'...")
    typer.echo(f"Writing updated dataset to {output_path}...")
    typer.echo(f"Max concurrent uploads: {max_concurrent_uploads}...")
    typer.echo(f"Base image directory: {base_image_dir}...")
    typer.echo(f"From key: {from_key}...")
    typer.echo(f"Skip upload: {skip_upload}...")
    typer.echo(f"Thinking level: {thinking_level}...")

    module = importlib.import_module(config_module)
    base_cls = getattr(module, "OUTPUT_SCHEMA")

    if not issubclass(base_cls, BaseModel):
        typer.echo(
            f"Error: {config_module}.OUTPUT_SCHEMA is not a subclass of pydantic.BaseModel"
        )
        raise typer.Exit(code=1)

    instructions = getattr(module, "INSTRUCTIONS", None) or None

    if instructions:
        typer.echo(f"Using instructions: '{instructions}'...")
    else:
        typer.echo("No instructions provided.")

    # JSON Schema is supoorted natively by Vertex AI and Gemini APIs,
    # but not yet on Batch Inference...
    # So we convert the JSON schema to Google internal "Schema"
    # google_json_schema = base_cls.model_json_schema()
    google_json_schema = convert_pydantic_model_to_google_schema(base_cls)
    asyncio.run(
        generate_batch_dataset(
            data_path=data_path,
            output_path=output_path,
            google_json_schema=google_json_schema,
            instructions=instructions,
            bucket_name=bucket_name,
            bucket_dir_name=bucket_dir_name,
            max_concurrent_uploads=max_concurrent_uploads,
            base_image_dir=base_image_dir,
            from_key=from_key,
            skip_upload=skip_upload,
            thinking_level=thinking_level,
        )
    )


@app.command(name="launch-batch-job")
def launch_batch_job_command(
    run_name: Annotated[str, typer.Argument(..., help="Name of the batch job run")],
    dataset_path: Annotated[Path, typer.Option(..., help="Path to the dataset file")],
    model: Annotated[str, typer.Option(..., help="Model to use for the batch job")],
    location: Annotated[
        str,
        typer.Option(..., help="GCP location where to run the batch job"),
    ] = "europe-west4",
):
    """Launch a Gemini Batch Inference job."""
    from labelr.google_genai import launch_batch_job

    launch_batch_job(
        run_name=run_name,
        dataset_path=dataset_path,
        model=model,
        location=location,
    )


@app.command()
def upload_training_dataset_from_predictions(
    prediction_path: Annotated[
        Path,
        typer.Argument(
            ...,
            help="Path to the prediction JSONL file generated by Google Inference Batch",
            exists=True,
            dir_okay=False,
            readable=True,
        ),
    ],
    instructions_path: Annotated[
        Path,
        typer.Option(
            ...,
            help="Path to the file with the instruction prompt for the model",
            exists=True,
            dir_okay=False,
            readable=True,
        ),
    ],
    json_schema_path: Annotated[
        Path,
        typer.Option(
            ...,
            help="Path to the file with the JSON schema to follow",
            dir_okay=False,
            readable=True,
        ),
    ],
    repo_id: Annotated[
        str, typer.Option(help="Hugging Face Datasets repository ID to push to")
    ],
    revision: Annotated[
        str,
        typer.Option(
            help="Revision (branch, tag or commit) to use for the Hugging Face Datasets repository"
        ),
    ] = "main",
    is_openfoodfacts_dataset: Annotated[
        bool, typer.Option(..., help="Whether this is an Open Food Facts dataset")
    ] = False,
    openfoodfacts_flavor: Annotated[
        Flavor,
        typer.Option(
            ...,
            help="Open Food Facts flavor of the dataset (if applicable)",
        ),
    ] = Flavor.off,
    split: Annotated[str, typer.Option(..., help="Name of the split")] = "train",
    tmp_dir: Annotated[
        Path | None,
        typer.Option(
            ...,
            help="Temporary directory to use for intermediate files, default to a temporary directory "
            "generated automatically. This is useful to relaunch the command if it fails midway.",
        ),
    ] = None,
    skip: Annotated[int, typer.Option(..., help="Number of samples to skip")] = 0,
    limit: Annotated[
        int | None,
        typer.Option(
            ..., help="Limit number of samples to process, or None for no limit"
        ),
    ] = None,
    raise_on_invalid_sample: Annotated[
        bool,
        typer.Option(
            ...,
            help="Whether to raise an error on invalid samples instead of skipping them",
        ),
    ] = False,
    image_max_size: Annotated[
        int | None,
        typer.Option(
            help="Maximum size (in pixels) for the images. If None, no resizing is performed.",
        ),
    ] = None,
):
    """Upload a training dataset to a Hugging Face Datasets repository from a
    Gemini batch prediction file."""
    import tempfile

    import orjson
    from huggingface_hub import HfApi

    from labelr.export.llm import export_to_hf_llm_image_extraction
    from labelr.google_genai import generate_sample_iter

    instructions = instructions_path.read_text()
    print(f"Instructions: {instructions}")
    json_schema = orjson.loads(json_schema_path.read_text())

    # We check that all fields are marked as required
    check_json_schema(json_schema)

    api = HfApi()
    config = {
        "instructions": instructions,
        "json_schema": json_schema,
    }
    with tempfile.TemporaryDirectory() as config_tmp_dir_str:
        config_tmp_dir = Path(config_tmp_dir_str)
        config_path = config_tmp_dir / "config.json"
        config_path.write_text(
            orjson.dumps(config, option=orjson.OPT_INDENT_2).decode("utf-8")
        )
        api.upload_file(
            path_or_fileobj=config_path,
            path_in_repo="config.json",
            repo_id=repo_id,
            repo_type="dataset",
        )
    sample_iter = generate_sample_iter(
        prediction_path=prediction_path,
        json_schema=json_schema,
        is_openfoodfacts_dataset=is_openfoodfacts_dataset,
        openfoodfacts_flavor=openfoodfacts_flavor,
        skip=skip,
        limit=limit,
        raise_on_invalid_sample=raise_on_invalid_sample,
    )
    export_to_hf_llm_image_extraction(
        sample_iter=sample_iter,
        split=split,
        repo_id=repo_id,
        revision=revision,
        tmp_dir=tmp_dir,
        image_max_size=image_max_size,
    )



import io
from pathlib import Path

from google.cloud import storage
from openfoodfacts.images import download_image as _download_image
from openfoodfacts.utils import ImageDownloadItem
from PIL import Image


def parse_hf_repo_id(hf_repo_id: str) -> tuple[str, str]:
    """Parse the repo_id and the revision from a hf_repo_id in the format:
    `org/repo-name@revision`.

    Returns a tuple (repo_id, revision), with revision = 'main' if it
    was not provided.
    """
    if "@" in hf_repo_id:
        hf_repo_id, revision = hf_repo_id.split("@", 1)
    else:
        revision = "main"

    return hf_repo_id, revision


def download_image(
    image: str | tuple[str, str],
    *,
    error_raise: bool = True,
    return_struct: bool = False,
    **kwargs,
) -> Image.Image | ImageDownloadItem | None:
    """Download an image from a URL or GCS URI and return it as a PIL Image.
    Args:
        image (str | tuple[str, str]): The URL or GCS URI of the image.
        error_raise (bool): Whether to raise an error if the image cannot be
            downloaded.
        return_struct (bool): Whether to return an ImageDownloadItem struct
            instead of a PIL Image.
        **kwargs: Additional arguments to pass to the download function.
    Returns:
        Image.Image | ImageDownloadItem: The downloaded image as a PIL Image
            or an ImageDownloadItem struct.
    """
    if isinstance(image, str) and image.startswith("gs://"):
        return download_image_from_gcs(image, return_struct=return_struct, **kwargs)
    return _download_image(
        image,
        error_raise=error_raise,
        return_struct=return_struct,
        **kwargs,
    )


def download_image_from_gcs(
    image_uri: str, client: storage.Client | None = None, return_struct: bool = False
) -> Image.Image | ImageDownloadItem:
    """Download an image from a Google Cloud Storage URI and return it as a
    PIL Image.

    Args:
        image_uri (str): The GCS URI of the image
            (e.g., gs://bucket_name/path/to/image.jpg).
        client (storage.Client | None): An optional Google Cloud Storage
            client. If not provided, a new client will be created.
    """
    if client is None:
        client = storage.Client()

    bucket_name, blob_name = image_uri.replace("gs://", "").split("/", 1)
    bucket = client.bucket(bucket_name)
    blob = bucket.blob(blob_name)
    image_data = blob.download_as_bytes()
    pil_image = Image.open(io.BytesIO(image_data))

    if return_struct:
        return ImageDownloadItem(
            url=image_uri,
            image=pil_image,
            error=None,
        )
    return pil_image


class PathWithContext:
    """A context manager that yields a Path object.

    This is useful to have a common interface with tempfile.TemporaryDirectory
    without actually creating a temporary directory.
    """

    def __init__(self, path: Path):
        self.path = path

    def __enter__(self) -> Path:
        return self.path

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        pass



import enum
import json
import typing
from pathlib import Path
from typing import Annotated

import typer
from openfoodfacts.utils import get_logger

from . import typer_description
from ...config import config, check_required_field
from .common import check_label_studio_api_key

app = typer.Typer(no_args_is_help=True)

logger = get_logger(__name__)


class PredictorBackend(enum.StrEnum):
    ultralytics = enum.auto()
    ultralytics_yolo_world = enum.auto()
    ultralytics_sam3 = enum.auto()


YOLO_WORLD_MODELS = (
    "yolov8s-world.pt",
    "yolov8s-worldv2.pt",
    "yolov8m-world.pt",
    "yolov8m-worldv2.pt",
    "yolov8l-world.pt",
    "yolov8l-worldv2.pt",
    "yolov8x-world.pt",
    "yolov8x-worldv2.pt",
)


@app.command()
def add(
    model_version: Annotated[
        str,
        typer.Option(
            help="Set the model version field of the prediction sent to Label Studio. "
            "This is used to track which model generated the prediction."
        ),
    ],
    project_id: Annotated[
        int | None, typer.Option(help=typer_description.LABEL_STUDIO_PROJECT_ID)
    ] = config.label_studio_project_id,
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    view_id: Annotated[
        int | None,
        typer.Option(help=typer_description.LABEL_STUDIO_VIEW_ID),
    ] = None,
    model_name: Annotated[
        str | None,
        typer.Option(
            "--model",
            help="Name or path of the object detection model to run. How this is used depends "
            "on the backend. If using `ultralytics` backend, the option is required and is the "
            "name of the model to download from the Ultralytics model zoo or the path to a local "
            "model. "
            "If using `ultralytics_yolo_world` backend, this is optional and is the name of the "
            "`yolo-world` model to download from the Ultralytics model zoo or the path to a local "
            "model (Defaults: `yolov8x-worldv2.pt`). "
            "If using `ultralytics_sam3` backend, this option is ignored, as there is a single model. "
            "The model is downloaded automatically from Hugging Face.",
        ),
    ] = None,
    skip_existing: Annotated[
        bool,
        typer.Option(
            help="Skip tasks that already have predictions",
        ),
    ] = True,
    backend: Annotated[
        PredictorBackend,
        typer.Option(
            help="The prediction backend, possible options are: `ultralytics` or `ultralytics_sam3`"
        ),
    ] = PredictorBackend.ultralytics,
    labels: Annotated[
        list[str] | None,
        typer.Option(
            help="List of class labels to use for Yolo model. If you're using Yolo-World or other "
            "zero-shot models, this is the list of label names that are going to be provided to the "
            "model. In such case, you can use `label_mapping` to map the model's output to the "
            "actual class names expected by Label Studio."
        ),
    ] = None,
    label_mapping: Annotated[
        str | None,
        typer.Option(
            help='Mapping of model labels to class names, as a JSON string. Example: \'{"price tag": "price-tag"}\''
        ),
    ] = None,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
    threshold: Annotated[
        float | None,
        typer.Option(
            help="Confidence threshold for selecting bounding boxes. The default is 0.1 for "
            "ultralytics backend and 0.25 for ultralytics_sam3 backend."
        ),
    ] = None,
    max_det: Annotated[int, typer.Option(help="Maximum numbers of detections")] = 300,
    dry_run: Annotated[
        bool,
        typer.Option(
            help="Launch in dry run mode, without uploading annotations to Label Studio"
        ),
    ] = False,
    error_raise: Annotated[
        bool,
        typer.Option(help="Raise an error if image download fails"),
    ] = True,
    imgsz: Annotated[
        int | None,
        typer.Option(
            help="Image size to use for Ultralytics models. If not provided, "
            "the default size of the model is used."
        ),
    ] = None,
):
    """Add predictions as pre-annotations to Label Studio tasks."""

    import tqdm
    from huggingface_hub import hf_hub_download
    from label_studio_sdk.client import LabelStudio
    from openfoodfacts.utils import get_image_from_url
    from PIL.ImageOps import exif_transpose
    from PIL import Image

    from labelr.annotate import format_annotation_results_from_ultralytics

    check_label_studio_api_key(api_key)
    check_required_field("--project-id", project_id)

    label_mapping_dict = None
    if label_mapping:
        label_mapping_dict = json.loads(label_mapping)

    if dry_run:
        logger.info("** Dry run mode enabled **")

    if backend == PredictorBackend.ultralytics and not Path(model_name).is_file():
        raise typer.BadParameter(
            f"Model file '{model_name}' not found. When the backend is `ultralytics` "
            "and the --model does not refer to a YOLO-World model, --model is expected "
            "to be a local Ultralytics model file (`.pt`)."
        )

    logger.info(
        "backend: %s, model_name: %s, labels: %s, threshold: %s, label mapping: %s, view ID: %s",
        backend,
        model_name,
        labels,
        threshold,
        label_mapping,
        view_id,
    )
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)

    if backend in (
        PredictorBackend.ultralytics,
        PredictorBackend.ultralytics_yolo_world,
    ):
        from ultralytics import YOLO, YOLOWorld

        if labels is None:
            raise typer.BadParameter("Labels are required for `ultralytics` backend")

        if threshold is None:
            threshold = 0.1

        if backend == PredictorBackend.ultralytics:
            model = YOLO(model_name)
        elif backend == PredictorBackend.ultralytics_yolo_world:
            if model_name is None:
                model_name = "yolov8x-worldv2.pt"
            model = YOLOWorld(model_name)
            model.set_classes(labels)

    elif backend == PredictorBackend.ultralytics_sam3:
        from ultralytics.models.sam import SAM3SemanticPredictor

        if threshold is None:
            threshold = 0.25

        # SAM3 cannot be downloaded directly using to to a gated access. Use a
        # proxy repo.
        model_path = hf_hub_download(
            "1038lab/sam3",
            filename="sam3.pt",
            revision="f055b060a4de0a040891ba2ebac9c5cb3c1c0132",
        )
        overrides = dict(
            task="segment",
            mode="predict",
            model=model_path,
            save=False,
        )

        if imgsz is not None:
            overrides["imgsz"] = imgsz
        model = SAM3SemanticPredictor(overrides=overrides)
    else:
        raise typer.BadParameter(f"Unsupported backend: {backend}")

    for task in tqdm.tqdm(
        ls.tasks.list(project=project_id, view=view_id), desc="tasks"
    ):
        if not (skip_existing and task.total_predictions > 0):
            image_url = task.data["image_url"]
            image = typing.cast(
                Image.Image | None,
                get_image_from_url(image_url, error_raise=error_raise),
            )
            if image is None:
                continue

            # Make sure that the image orientation is in accordance with the EXIF metadata
            exif_transpose(image, in_place=True)
            min_score = None
            if backend in (
                PredictorBackend.ultralytics,
                PredictorBackend.ultralytics_yolo_world,
            ):
                predict_kwargs = {
                    "conf": threshold,
                    "max_det": max_det,
                }
                if imgsz is not None:
                    predict_kwargs["imgsz"] = imgsz
                results = model.predict(image, **predict_kwargs)[0]
                labels = typing.cast(list[str], labels)
                label_studio_result = format_annotation_results_from_ultralytics(
                    results, labels, label_mapping_dict
                )
                min_score = min(results.boxes.conf.tolist(), default=None)
            elif backend == PredictorBackend.ultralytics_sam3:
                model.set_image(image)
                results = model(text=labels)[0]
                label_studio_result = format_annotation_results_from_ultralytics(
                    results, labels, label_mapping_dict
                )
                min_score = min(results.boxes.conf.tolist(), default=None)
            if dry_run:
                logger.info("image_url: %s", image_url)
                logger.info("result: %s", label_studio_result)
            else:
                ls.predictions.create(
                    task=task.id,
                    result=label_studio_result,
                    model_version=model_version,
                    score=min_score,
                )
                logger.info("Prediction added for task: %s", task.id)


@app.command()
def delete(
    model_version: Annotated[
        str,
        typer.Option(
            help="The model version associated with the prediction set to delete"
        ),
    ],
    project_id: Annotated[
        int | None, typer.Option(help=typer_description.LABEL_STUDIO_PROJECT_ID)
    ] = config.label_studio_project_id,
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
    dry_run: Annotated[
        bool,
        typer.Option(help="Launch in dry run mode, without deleting any prediction"),
    ] = False,
):
    """Delete predictions for a given model version in a Label Studio project."""
    from label_studio_sdk.client import LabelStudio
    import tqdm

    check_label_studio_api_key(api_key)
    check_required_field("--project-id", project_id)
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)

    query = {
        "conjunction": "and",
        "items": [
            {
                "filter": "filter:tasks:predictions_model_versions",
                "operator": "contains",
                "value": [model_version],
                "type": "List",
            }
        ],
    }

    for task in tqdm.tqdm(ls.tasks.list(project=project_id, query=query), desc="tasks"):
        for prediction in task.predictions:
            if prediction["model_version"] != model_version:
                continue
            if dry_run:
                logger.info("Dry run: prediction %s would be deleted", prediction["id"])
            else:
                ls.predictions.delete(prediction["id"])
                logger.info("Prediction deleted: %s", prediction["id"])


@app.command()
def show_uncertain(
    model_version: Annotated[
        str,
        typer.Option(help="The model version associated with the prediction"),
    ],
    threshold: Annotated[float, typer.Option(help="Score threshold")],
    add_score: Annotated[
        bool,
        typer.Option(
            help="Whether to add the score as a second column in output, making output a CSV file. "
            "This option is ignored if --output is not provided."
        ),
    ] = True,
    output: Annotated[
        str | None,
        typer.Option(help="Path to an file where the task IDs that match (optional)"),
    ] = None,
    project_id: Annotated[
        int | None, typer.Option(help=typer_description.LABEL_STUDIO_PROJECT_ID)
    ] = config.label_studio_project_id,
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
):
    """Show all tasks that have a prediction named `model_version` with a minimum
    score below `threshold`."""
    from label_studio_sdk.client import LabelStudio
    import tqdm
    import sys

    check_label_studio_api_key(api_key)
    check_required_field("--project-id", project_id)
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)

    if output == "-":
        f = sys.stdout
    elif output is None:
        f = open("/dev/null", "w")
    else:
        f = open(output, "w")

    with f:
        for task in tqdm.tqdm(
            ls.tasks.list(project=project_id, fields="all"), desc="tasks"
        ):
            for prediction in task.predictions:
                if prediction["model_version"] != model_version:
                    continue
                elif (min_score := (prediction["score"] or 0.0)) <= threshold:
                    logger.info("Task ID: %s, score: %s", task.id, min_score)
                    line = f"{task.id},{min_score}\n" if add_score else f"{task.id}\n"
                    f.write(line)
                    f.flush()


@app.command()
def annotate_from(
    project_id: Annotated[
        int | None, typer.Option(help=typer_description.LABEL_STUDIO_PROJECT_ID)
    ] = config.label_studio_project_id,
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    updated_by: Annotated[
        int | None, typer.Option(help="User ID to declare as annotator")
    ] = None,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
):
    """Create annotations for all tasks from predictions.

    This command is useful if you imported tasks with predictions, and want to
    "validate" these predictions by creating annotations.
    """
    import tqdm
    from label_studio_sdk.client import LabelStudio
    from label_studio_sdk.types import LseTask

    check_label_studio_api_key(api_key)
    check_required_field("--project-id", project_id)
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)

    task: LseTask
    for task in tqdm.tqdm(
        ls.tasks.list(project=project_id, fields="all"), desc="tasks"
    ):
        task_id = task.id
        if task.total_annotations == 0 and task.total_predictions > 0:
            logger.info("Creating annotation for task: %s", task_id)
            ls.annotations.create(
                id=task_id,
                result=task.predictions[0]["result"],
                project=project_id,
                updated_by=updated_by,
            )



import logging
import random
import string
import typing

import datasets
from openfoodfacts import Flavor
from openfoodfacts.barcode import normalize_barcode
from openfoodfacts.images import download_image, generate_image_url
from PIL import Image, ImageOps

logger = logging.getLogger(__name__)


def format_annotation_results_from_hf_to_ls(
    objects: dict, image_width: int, image_height: int
):
    """Format annotation results from a HF object detection dataset into Label
    Studio format."""
    annotation_results = []
    for i in range(len(objects["bbox"])):
        bbox = objects["bbox"][i]
        # category_id = objects["category_id"][i]
        category_name = objects["category_name"][i]
        # These are relative coordinates (between 0.0 and 1.0)
        y_min, x_min, y_max, x_max = bbox
        # Make sure the coordinates are within the image boundaries,
        # and convert them to percentages
        y_min = min(max(0, y_min), 1.0) * 100
        x_min = min(max(0, x_min), 1.0) * 100
        y_max = min(max(0, y_max), 1.0) * 100
        x_max = min(max(0, x_max), 1.0) * 100
        x = x_min
        y = y_min
        width = x_max - x_min
        height = y_max - y_min

        id_ = "".join(random.choices(string.ascii_letters + string.digits, k=10))
        annotation_results.append(
            {
                "id": id_,
                "type": "rectanglelabels",
                "from_name": "label",
                "to_name": "image",
                "original_width": image_width,
                "original_height": image_height,
                "image_rotation": 0,
                "value": {
                    "rotation": 0,
                    "x": x,
                    "y": y,
                    "width": width,
                    "height": height,
                    "rectanglelabels": [category_name],
                },
            },
        )
    return annotation_results


def format_object_detection_sample_from_hf_to_ls(hf_sample: dict, split: str) -> dict:
    hf_meta = hf_sample["meta"]
    objects = hf_sample["objects"]
    image_width = hf_sample["width"]
    image_height = hf_sample["height"]
    annotation_results = format_annotation_results_from_hf_to_ls(
        objects, image_width, image_height
    )
    image_id = hf_sample["image_id"]
    image_url = hf_meta["image_url"]
    meta_kwargs = {}

    if "off_image_id" in hf_meta:
        # If `off_image_id` is present, we assume this is an Open Food Facts
        # dataset sample.
        # We normalize the barcode, and generate a new image URL
        # to make sure that:
        # - the image URL is valid with correct path
        # - we use the images subdomain everywhere
        off_image_id = hf_meta["off_image_id"]
        meta_kwargs["off_image_id"] = off_image_id
        barcode = normalize_barcode(hf_meta["barcode"])
        meta_kwargs["barcode"] = barcode
        image_id = f"{barcode}_{off_image_id}"

        if ".openfoodfacts." in image_url:
            flavor = Flavor.off
        elif ".openbeautyfacts." in image_url:
            flavor = Flavor.obf
        elif ".openpetfoodfacts." in image_url:
            flavor = Flavor.opf
        elif ".openproductsfacts." in image_url:
            flavor = Flavor.opf
        else:
            raise ValueError(
                f"Unknown Open Food Facts flavor for image URL: {image_url}"
            )
        image_url = generate_image_url(
            code=barcode, image_id=off_image_id, flavor=flavor
        )

    return {
        "data": {
            "image_id": image_id,
            "image_url": image_url,
            "batch": "null",
            "split": split,
            "meta": {
                "width": image_width,
                "height": image_height,
                **meta_kwargs,
            },
        },
        "predictions": [{"result": annotation_results}],
    }


def format_object_detection_sample_to_ls(
    image_id: str,
    image_url: str,
    width: int,
    height: int,
    extra_meta: dict | None = None,
) -> dict:
    """Format an object detection sample in Label Studio format.

    Args:
        image_id: The image ID.
        image_url: The URL of the image.
        width: The width of the image.
        height: The height of the image.
        extra_meta: Extra metadata to include in the sample.
    """
    extra_meta = extra_meta or {}
    return {
        "data": {
            "image_id": image_id,
            "image_url": image_url,
            "batch": "null",
            "meta": {
                "width": width,
                "height": height,
                **extra_meta,
            },
        },
    }


def format_object_detection_sample_to_hf(
    task_data: dict,
    annotations: list[dict],
    label_names: list[str],
    merge_labels: bool = False,
    use_aws_cache: bool = False,
    image_max_size: int | None = None,
    skip_labels: list[str] | None = None,
) -> dict | None:
    """Format a Label Studio object detection sample to Hugging Face format.

    Args:
        task_data: The task data from Label Studio.
        annotations: The annotations from Label Studio.
        label_names: The list of label names.
        merge_labels: Whether to merge all labels into a single label (the
            first label in `label_names`).
        use_aws_cache: Whether to use AWS cache when downloading images.
        image_max_size: Maximum size (in pixels) for the images.
            If None, no resizing is performed. Defaults to None.
        skip_labels (list[str] | None): List of label names to skip. If the
            label of an annotation result is in this list, it will be skipped.
            Defaults to None.

    Returns:
        The formatted sample, or None in the following cases:
        - More than one annotation is found
        - No annotation is found
        - An error occurs when downloading the image
    """
    if len(annotations) > 1:
        logger.info("More than one annotation found, skipping")
        return None
    elif len(annotations) == 0:
        logger.info("No annotation found, skipping")
        return None

    skip_labels = skip_labels or []
    annotation = annotations[0]
    bboxes = []
    bbox_label_ids = []
    bbox_label_names = []

    for annotation_result in annotation["result"]:
        if annotation_result["type"] != "rectanglelabels":
            continue
            # raise ValueError("Invalid annotation type: %s" % annotation_result["type"])

        value = annotation_result["value"]
        x_min = value["x"] / 100
        y_min = value["y"] / 100
        width = value["width"] / 100
        height = value["height"] / 100
        x_max = x_min + width
        y_max = y_min + height
        bboxes.append([y_min, x_min, y_max, x_max])

        label_name = label_names[0] if merge_labels else value["rectanglelabels"][0]
        if label_name in skip_labels:
            continue
        label_id = label_names.index(label_name) if label_name in label_names else -1
        if label_id == -1:
            logger.warning(
                "Label name '%s' not found in label_names, skipping annotation result: '%s'",
                label_name,
                annotation_result,
            )
            continue
        bbox_label_names.append(label_name)
        bbox_label_ids.append(label_id)

    image_url = task_data["image_url"]
    image = download_image(image_url, error_raise=False, use_cache=use_aws_cache)
    if image is None:
        logger.error("Failed to download image: %s", image_url)
        return None

    image = typing.cast(Image.Image, image)
    # Correct image orientation using EXIF data
    # Label Studio provides bounding boxes based on the displayed image (after
    # eventual EXIF rotation), so we need to apply the same transformation to
    # the image.
    # Indeed, Hugging Face stores images without applying EXIF rotation, and
    # EXIF data is not preserved in the dataset.
    ImageOps.exif_transpose(image, in_place=True)
    if image.mode != "RGB":
        # Convert image to RGB if needed
        image = image.convert("RGB")
    # Resize image if larger than max size
    if image_max_size is not None and (
        image.width > image_max_size or image.height > image_max_size
    ):
        image.thumbnail((image_max_size, image_max_size), Image.Resampling.LANCZOS)

    meta = task_data.get("meta", {})
    barcode = meta.get("barcode", None)
    off_image_id = meta.get("off_image_id", None)
    width = image.width
    height = image.height
    return {
        "image_id": task_data["image_id"],
        "image": image,
        "width": width,
        "height": height,
        "meta": {
            "barcode": barcode,
            "off_image_id": off_image_id,
            "image_url": image_url,
        },
        "objects": {
            "bbox": bboxes,
            "category_id": bbox_label_ids,
            "category_name": bbox_label_names,
        },
    }


def get_hf_object_detection_features(
    is_openfoodfacts_dataset: bool,
) -> datasets.Features:
    """Get the HuggingFace Dataset features for object detection.

    Args:
        is_openfoodfacts_dataset (bool): Whether the dataset is an Open Food
            Facts dataset. If True, the dataset will include additional
            metadata fields specific to Open Food Facts (`barcode` and
            `off_image_id`).
    """
    features_dict = {
        "image_id": datasets.Value("string"),
        "image": datasets.features.Image(),
        "width": datasets.Value("int64"),
        "height": datasets.Value("int64"),
        "meta": {
            "image_url": datasets.Value("string"),
        },
        "objects": {
            "bbox": datasets.Sequence(datasets.Sequence(datasets.Value("float32"))),
            "category_id": datasets.Sequence(datasets.Value("int64")),
            "category_name": datasets.Sequence(datasets.Value("string")),
        },
    }

    if is_openfoodfacts_dataset:
        features_dict["meta"]["barcode"] = datasets.Value("string")
        features_dict["meta"]["off_image_id"] = datasets.Value("string")

    return datasets.Features(features_dict)



import functools
import logging
import pickle
import random
import tempfile
import typing
from pathlib import Path

import datasets
import tqdm
from label_studio_sdk.client import LabelStudio
from openfoodfacts.images import download_image
from PIL import Image, ImageOps

from labelr.export.common import _pickle_sample_generator
from labelr.sample.object_detection import (
    format_object_detection_sample_to_hf,
    get_hf_object_detection_features,
)

logger = logging.getLogger(__name__)


def export_from_ls_to_hf_object_detection(
    ls: LabelStudio,
    repo_id: str,
    label_names: list[str],
    project_id: int,
    is_openfoodfacts_dataset: bool,
    image_max_size: int | None = None,
    view_id: int | None = None,
    merge_labels: bool = False,
    use_aws_cache: bool = True,
    revision: str = "main",
    skip_labels: list[str] | None = None,
) -> None:
    """Export annotations from a Label Studio project to a Hugging Face
    dataset.

    The Label Studio project should be an object detection project.

    Args:
        ls (LabelStudio): Label Studio client instance.
        repo_id (str): Hugging Face repository ID to push the dataset to.
        label_names (list[str]): List of label names in the project.
        project_id (int): Label Studio project ID to export from.
        is_openfoodfacts_dataset (bool): Whether the dataset is an Open Food
            Facts dataset. If True, the dataset will include additional
            metadata fields specific to Open Food Facts (`barcode` and
            `off_image_id`).
        image_max_size (int | None): Maximum size (in pixels) for the images.
            If None, no resizing is performed. Defaults to None.
        view_id (int | None): Label Studio view ID to export from. If None,
            all tasks are exported. Defaults to None.
        merge_labels (bool): Whether to merge all labels into a single label
            named "object". Defaults to False.
        use_aws_cache (bool): Whether to use the AWS image cache when
            downloading images. Defaults to True.
        revision (str): The dataset revision to push to. Defaults to 'main'.
        skip_labels (list[str] | None): List of label names to skip. If the
            label of an annotation result is in this list, it will be skipped.
            Defaults to None.
    """
    if merge_labels:
        label_names = ["object"]

    logger.info(
        "Project ID: %d, label names: %s, skip labels: %s, repo_id: %s, revision: %s, view ID: %s",
        project_id,
        label_names,
        skip_labels,
        repo_id,
        revision,
        view_id,
    )

    for split in ["train", "val"]:
        logger.info("Processing split: %s", split)

        with tempfile.TemporaryDirectory() as tmp_dir_str:
            tmp_dir = Path(tmp_dir_str)
            logger.info("Saving samples to temporary directory: %s", tmp_dir)
            for i, task in tqdm.tqdm(
                enumerate(
                    ls.tasks.list(project=project_id, fields="all", view=view_id)
                ),
                desc="tasks",
            ):
                if task.data["split"] != split:
                    continue
                sample = format_object_detection_sample_to_hf(
                    task_data=task.data,
                    annotations=task.annotations,
                    label_names=label_names,
                    merge_labels=merge_labels,
                    use_aws_cache=use_aws_cache,
                    image_max_size=image_max_size,
                    skip_labels=skip_labels,
                )
                if sample is not None:
                    # Save output as pickle
                    with open(tmp_dir / f"{split}_{i:05}.pkl", "wb") as f:
                        pickle.dump(sample, f)

            features = get_hf_object_detection_features(is_openfoodfacts_dataset)
            hf_ds = datasets.Dataset.from_generator(
                functools.partial(_pickle_sample_generator, tmp_dir),
                features=features,
            )
            hf_ds.push_to_hub(repo_id, split=split, revision=revision)


def export_from_ls_to_ultralytics_object_detection(
    ls: LabelStudio,
    output_dir: Path,
    label_names: list[str],
    project_id: int,
    train_ratio: float = 0.8,
    error_raise: bool = True,
    merge_labels: bool = False,
    use_aws_cache: bool = True,
    view_id: int | None = None,
    image_max_size: int | None = None,
    skip_labels: list[str] | None = None,
):
    """Export annotations from a Label Studio project to the Ultralytics
    format.

    The Label Studio project should be an object detection project with a
    single rectanglelabels annotation result per task.

    Args:
        ls (LabelStudio): Label Studio client instance.
        output_dir (Path): Path to the output directory.
        label_names (list[str]): List of label names in the project.
        project_id (int): Label Studio project ID to export from.
        train_ratio (float): Ratio of training samples. The rest will be used
            for validation. Defaults to 0.8.
        error_raise (bool): Whether to raise an error if an image fails to
            download. If False, the image will be skipped. Defaults to True.
        merge_labels (bool): Whether to merge all labels into a single label
            named "object". Defaults to False.
        use_aws_cache (bool): Whether to use the AWS image cache when
            downloading images. Defaults to True.
        view_id (int | None): Label Studio view ID to export from. If None,
            all tasks are exported. Defaults to None.
        image_max_size (int | None): Maximum size (in pixels) for the images.
            If None, no resizing is performed. Defaults to None.
        skip_labels (list[str] | None): List of label names to skip. If the
            label of an annotation result is in this list, it will be skipped.
            Defaults to None.
    """
    if merge_labels:
        label_names = ["object"]

    skip_labels = skip_labels or []
    logger.info(
        "Project ID: %d, label names: %s, skip labels: %s",
        project_id,
        label_names,
        skip_labels,
    )

    data_dir = output_dir / "data"
    data_dir.mkdir(parents=True, exist_ok=True)
    split_warning_displayed = False

    # NOTE: before, all images were sent to val, the last split
    label_dir = data_dir / "labels"
    images_dir = data_dir / "images"
    for split in ["train", "val"]:
        (label_dir / split).mkdir(parents=True, exist_ok=True)
        (images_dir / split).mkdir(parents=True, exist_ok=True)

    for task in tqdm.tqdm(
        ls.tasks.list(project=project_id, fields="all", view=view_id),
        desc="tasks",
    ):
        split = task.data.get("split")

        if split is None:
            if not split_warning_displayed:
                logger.warning(
                    "Split information not found, assigning randomly. "
                    "To avoid this, set the `split` field in the task data."
                )
                split_warning_displayed = True
            split = "train" if random.random() < train_ratio else "val"

        elif split not in ["train", "val"]:
            raise ValueError("Invalid split name: %s", split)

        if len(task.annotations) > 1:
            logger.warning("More than one annotation found, skipping")
            continue
        elif len(task.annotations) == 0:
            logger.debug("No annotation found, skipping")
            continue

        annotation = task.annotations[0]
        if annotation["was_cancelled"] is True:
            logger.debug("Annotation was cancelled, skipping")
            continue

        if "image_id" not in task.data:
            raise ValueError(
                "`image_id` field not found in task data. "
                "Make sure the task data contains the `image_id` "
                "field, which should be a unique identifier for the image."
            )
        if "image_url" not in task.data:
            raise ValueError(
                "`image_url` field not found in task data. "
                "Make sure the task data contains the `image_url` "
                "field, which should be the URL of the image."
            )
        image_id = task.data["image_id"]
        image_url = task.data["image_url"]

        has_valid_annotation = False
        with (label_dir / split / f"{image_id}.txt").open("w") as f:
            if not any(
                annotation_result["type"] == "rectanglelabels"
                for annotation_result in annotation["result"]
            ):
                continue

            for annotation_result in annotation["result"]:
                if annotation_result["type"] == "rectanglelabels":
                    value = annotation_result["value"]
                    x_min = value["x"] / 100
                    y_min = value["y"] / 100
                    width = value["width"] / 100
                    height = value["height"] / 100
                    label_name = (
                        label_names[0] if merge_labels else value["rectanglelabels"][0]
                    )
                    if label_name in skip_labels:
                        continue
                    label_id = (
                        label_names.index(label_name)
                        if label_name in label_names
                        else -1
                    )
                    if label_id == -1:
                        logger.warning(
                            "Label name '%s' not found in label_names, skipping annotation result: '%s'",
                            label_name,
                            annotation_result,
                        )
                        continue

                    # Save the labels in the Ultralytics format:
                    # - one label per line
                    # - each line is a list of 5 elements:
                    #   - label_id
                    #   - x_center
                    #   - y_center
                    #   - width
                    #   - height
                    x_center = x_min + width / 2
                    y_center = y_min + height / 2
                    f.write(f"{label_id} {x_center} {y_center} {width} {height}\n")
                    has_valid_annotation = True

        if has_valid_annotation:
            image = download_image(
                image_url,
                return_struct=False,
                error_raise=error_raise,
                use_cache=use_aws_cache,
            )
            if image is None:
                logger.error("Failed to download image: %s", image_url)
                continue

            image = typing.cast(Image.Image, image)

            # Rotate image according to exif orientation using Pillow
            ImageOps.exif_transpose(image, in_place=True)
            # Resize image if larger than max size
            if image_max_size is not None and (
                image.width > image_max_size or image.height > image_max_size
            ):
                image.thumbnail(
                    (image_max_size, image_max_size), Image.Resampling.LANCZOS
                )
            image.save(images_dir / split / f"{image_id}.jpg", format="JPEG")

    with (output_dir / "data.yaml").open("w") as f:
        f.write("path: data\n")
        f.write("train: images/train\n")
        f.write("val: images/val\n")
        f.write("test:\n")
        f.write("names:\n")
        for i, label_name in enumerate(label_names):
            f.write(f"  {i}: {label_name}\n")


def export_from_hf_to_ultralytics_object_detection(
    repo_id: str,
    output_dir: Path,
    download_images: bool = True,
    error_raise: bool = True,
    use_aws_cache: bool = True,
    image_max_size: int | None = None,
    revision: str = "main",
    skip_labels: list[str] | None = None,
):
    """Export annotations from a Hugging Face dataset project to the
    Ultralytics format.

    The Label Studio project should be an object detection project with a
    single rectanglelabels annotation result per task.

    Args:
        repo_id (str): Hugging Face repository ID to load the dataset from.
        output_dir (Path): Path to the output directory.
        download_images (bool): Whether to download images from URLs in the
            dataset. If False, the dataset is expected to contain an `image`
            field with the image data.
        error_raise (bool): Whether to raise an error if an image fails to
            download. If False, the image will be skipped. This option is only
            used if `download_images` is True. Defaults to True.
        use_aws_cache (bool): Whether to use the AWS image cache when
            downloading images. This option is only used if `download_images`
            is True. Defaults to True.
        image_max_size (int | None): Maximum size (in pixels) for the images.
            If None, no resizing is performed. Defaults to None.
        revision (str): The dataset revision to load. Defaults to 'main'.
        skip_labels (list[str] | None): List of label names to skip. If the
            label of an annotation result is in this list, it will be skipped.
            Defaults to None.
    """
    skip_labels = skip_labels or []
    logger.info(
        "Repo ID: %s, revision: %s, skip labels: %s", repo_id, revision, skip_labels
    )
    ds = datasets.load_dataset(repo_id, revision=revision)
    data_dir = output_dir / "data"
    data_dir.mkdir(parents=True, exist_ok=True)
    label_id_to_name = {}

    split_map = {
        "train": "train",
        "val": "val",
    }
    if "val" not in ds and "test" in ds:
        logger.info("val split not found, using test split instead as val")
        split_map["val"] = "test"

    for split in ["train", "val"]:
        split_target = split_map[split]
        split_labels_dir = data_dir / "labels" / split
        split_labels_dir.mkdir(parents=True, exist_ok=True)
        split_images_dir = data_dir / "images" / split
        split_images_dir.mkdir(parents=True, exist_ok=True)

        for sample in tqdm.tqdm(ds[split_target], desc="samples"):
            image_id = sample["image_id"]

            if download_images:
                if "meta" not in sample or "image_url" not in sample["meta"]:
                    raise ValueError(
                        "`meta.image_url` field not found in sample. "
                        "Make sure the dataset contains the `meta.image_url` "
                        "field, which should be the URL of the image, or set "
                        "`download_images` to False."
                    )
                image_url = sample["meta"]["image_url"]
                image = download_image(
                    image_url,
                    return_struct=False,
                    error_raise=error_raise,
                    use_cache=use_aws_cache,
                )
                if image is None:
                    logger.error("Failed to download image: %s", image_url)
                    continue
            else:
                image = sample["image"]

            image = typing.cast(Image.Image, image)
            # Rotate image according to exif orientation using Pillow
            # If the image source is Hugging Face, EXIF data is not preserved,
            # so this step is only useful when downloading images.
            ImageOps.exif_transpose(image, in_place=True)
            if image.mode != "RGB":
                # Convert image to RGB if needed
                image = image.convert("RGB")
            # Resize image if larger than max size
            if image_max_size is not None and (
                image.width > image_max_size or image.height > image_max_size
            ):
                image.thumbnail(
                    (image_max_size, image_max_size), Image.Resampling.LANCZOS
                )
            image.save(split_images_dir / f"{image_id}.jpg")

            objects = sample["objects"]
            bboxes = objects["bbox"]
            label_ids = objects["category_id"]
            label_names = objects["category_name"]

            with (split_labels_dir / f"{image_id}.txt").open("w") as f:
                for bbox, label_id, label_name in zip(bboxes, label_ids, label_names):
                    if label_id not in label_id_to_name:
                        label_id_to_name[label_id] = label_name
                    if label_name in skip_labels:
                        continue
                    y_min, x_min, y_max, x_max = bbox
                    y_min = min(max(y_min, 0.0), 1.0)
                    x_min = min(max(x_min, 0.0), 1.0)
                    y_max = min(max(y_max, 0.0), 1.0)
                    x_max = min(max(x_max, 0.0), 1.0)
                    width = x_max - x_min
                    height = y_max - y_min
                    # Save the labels in the Ultralytics format:
                    # - one label per line
                    # - each line is a list of 5 elements:
                    #   - label_id
                    #   - x_center
                    #   - y_center
                    #   - width
                    #   - height
                    x_center = x_min + width / 2
                    y_center = y_min + height / 2
                    f.write(f"{label_id} {x_center} {y_center} {width} {height}\n")

    label_names = [x[1] for x in sorted(label_id_to_name.items(), key=lambda x: x[0])]
    with (output_dir / "data.yaml").open("w") as f:
        f.write("path: data\n")
        f.write("train: images/train\n")
        f.write("val: images/val\n")
        f.write("test:\n")
        f.write("names:\n")
        for i, label_name in enumerate(label_names):
            f.write(f"  {i}: {label_name}\n")



import tempfile
from pathlib import Path

import datasets
import fiftyone as fo
from huggingface_hub import hf_hub_download

from labelr.dataset_features import OBJECT_DETECTION_DS_PREDICTION_FEATURES
from labelr.utils import parse_hf_repo_id


def convert_bbox_to_fo_format(
    bbox: tuple[float, float, float, float],
) -> tuple[float, float, float, float]:
    # Bounding box coordinates should be relative values
    # in [0, 1] in the following format:
    # [top-left-x, top-left-y, width, height]
    y_min, x_min, y_max, x_max = bbox
    return (
        x_min,
        y_min,
        (x_max - x_min),
        (y_max - y_min),
    )


def visualize(
    hf_repo_id: str,
    dataset_name: str,
    persistent: bool,
):
    hf_repo_id, hf_revision = parse_hf_repo_id(hf_repo_id)

    file_path = hf_hub_download(
        hf_repo_id,
        filename="predictions.parquet",
        revision=hf_revision,
        repo_type="model",
        # local_dir="./predictions/",
    )
    file_path = Path(file_path).absolute()
    prediction_dataset = datasets.load_dataset(
        "parquet",
        data_files=str(file_path),
        split="train",
        features=OBJECT_DETECTION_DS_PREDICTION_FEATURES,
    )
    fo_dataset = fo.Dataset(name=dataset_name, persistent=persistent)

    with tempfile.TemporaryDirectory() as tmpdir_str:
        tmp_dir = Path(tmpdir_str)
        for i, hf_sample in enumerate(prediction_dataset):
            image = hf_sample["image"]
            image_path = tmp_dir / f"{i}.jpg"
            image.save(image_path)
            split = hf_sample["split"]
            sample = fo.Sample(
                filepath=image_path,
                split=split,
                tags=[split],
                image=hf_sample["image_id"],
            )
            ground_truth_detections = [
                fo.Detection(
                    label=hf_sample["objects"]["category_name"][i],
                    bounding_box=convert_bbox_to_fo_format(
                        bbox=hf_sample["objects"]["bbox"][i],
                    ),
                )
                for i in range(len(hf_sample["objects"]["bbox"]))
            ]
            sample["ground_truth"] = fo.Detections(detections=ground_truth_detections)

            if hf_sample["detected"] is not None and hf_sample["detected"]["bbox"]:
                model_detections = [
                    fo.Detection(
                        label=hf_sample["detected"]["category_name"][i],
                        bounding_box=convert_bbox_to_fo_format(
                            bbox=hf_sample["detected"]["bbox"][i]
                        ),
                        confidence=hf_sample["detected"]["confidence"][i],
                    )
                    for i in range(len(hf_sample["detected"]["bbox"]))
                ]
                sample["model"] = fo.Detections(detections=model_detections)

            fo_dataset.add_sample(sample)

        # View summary info about the dataset
        print(fo_dataset)

        # Print the first few samples in the dataset
        print(fo_dataset.head())

        # Visualize the dataset in the FiftyOne App
        session = fo.launch_app(fo_dataset)
        fo_dataset.evaluate_detections(
            "model", gt_field="ground_truth", eval_key="eval", compute_mAP=True
        )
        session.wait()



import functools
import logging
import pickle
import tempfile
from pathlib import Path

import datasets
from openfoodfacts.images import generate_image_url
from openfoodfacts.types import Flavor
from PIL import Image, ImageOps

from labelr.export.common import _pickle_sample_generator
from labelr.sample.classification import HF_DS_CLASSIFICATION_FEATURES

logger = logging.getLogger(__name__)


def export_from_ultralytics_to_hf_classification(
    dataset_dir: Path,
    repo_id: str,
    label_names: list[str],
    merge_labels: bool = False,
    is_openfoodfacts_dataset: bool = False,
    openfoodfacts_flavor: Flavor = Flavor.off,
) -> None:
    """Export an Ultralytics classification dataset to a Hugging Face dataset.

    The Ultralytics dataset directory should contain 'train', 'val' and/or
    'test' subdirectories, each containing subdirectories for each label.

    Args:
        dataset_dir (Path): Path to the Ultralytics dataset directory.
        repo_id (str): Hugging Face repository ID to push the dataset to.
        label_names (list[str]): List of label names.
        merge_labels (bool): Whether to merge all labels into a single label
            named 'object'.
        is_openfoodfacts_dataset (bool): Whether the dataset is from
            Open Food Facts. If True, the `off_image_id` and `image_url` will
            be generated automatically. `off_image_id` is extracted from the
            image filename.
        openfoodfacts_flavor (Flavor): Flavor of Open Food Facts dataset. This
            is ignored if `is_openfoodfacts_dataset` is False.
    """
    logger.info("Repo ID: %s, dataset_dir: %s", repo_id, dataset_dir)

    if not any((dataset_dir / split).is_dir() for split in ["train", "val", "test"]):
        raise ValueError(
            f"Dataset directory {dataset_dir} does not contain 'train', 'val' or 'test' subdirectories"
        )

    # Save output as pickle
    for split in ["train", "val", "test"]:
        split_dir = dataset_dir / split

        if not split_dir.is_dir():
            logger.info("Skipping missing split directory: %s", split_dir)
            continue

        with tempfile.TemporaryDirectory() as tmp_dir_str:
            tmp_dir = Path(tmp_dir_str)
            for label_dir in (d for d in split_dir.iterdir() if d.is_dir()):
                label_name = label_dir.name
                if merge_labels:
                    label_name = "object"
                if label_name not in label_names:
                    raise ValueError(
                        "Label name %s not in provided label names (label names: %s)"
                        % (label_name, label_names),
                    )
                label_id = label_names.index(label_name)

                for image_path in label_dir.glob("*"):
                    if is_openfoodfacts_dataset:
                        image_stem_parts = image_path.stem.split("_")
                        barcode = image_stem_parts[0]
                        off_image_id = image_stem_parts[1]
                        image_id = f"{barcode}_{off_image_id}"
                        image_url = generate_image_url(
                            barcode, off_image_id, flavor=openfoodfacts_flavor
                        )
                    else:
                        image_id = image_path.stem
                        barcode = ""
                        off_image_id = ""
                        image_url = ""
                    image = Image.open(image_path)
                    image.load()

                    if image.mode != "RGB":
                        image = image.convert("RGB")

                    # Rotate image according to exif orientation using Pillow
                    ImageOps.exif_transpose(image, in_place=True)
                    sample = {
                        "image_id": image_id,
                        "image": image,
                        "width": image.width,
                        "height": image.height,
                        "meta": {
                            "barcode": barcode,
                            "off_image_id": off_image_id,
                            "image_url": image_url,
                        },
                        "category_id": label_id,
                        "category_name": label_name,
                    }
                    with open(tmp_dir / f"{split}_{image_id}.pkl", "wb") as f:
                        pickle.dump(sample, f)

            hf_ds = datasets.Dataset.from_generator(
                functools.partial(_pickle_sample_generator, tmp_dir),
                features=HF_DS_CLASSIFICATION_FEATURES,
            )
            hf_ds.push_to_hub(repo_id, split=split)



import pickle
from pathlib import Path

from openfoodfacts.types import Flavor

from labelr.types import TaskType


def _pickle_sample_generator(dir: Path):
    """Generator that yields samples from pickles in a directory."""
    for pkl in dir.glob("*.pkl"):
        with open(pkl, "rb") as f:
            yield pickle.load(f)


def export_from_ultralytics_to_hf(
    task_type: TaskType,
    dataset_dir: Path,
    repo_id: str,
    label_names: list[str],
    merge_labels: bool = False,
    is_openfoodfacts_dataset: bool = False,
    openfoodfacts_flavor: Flavor = Flavor.off,
) -> None:
    from labelr.export.classification import (
        export_from_ultralytics_to_hf_classification,
    )

    if task_type != TaskType.classification:
        raise NotImplementedError(
            "Only classification task is currently supported for Ultralytics to HF export"
        )

    if task_type == TaskType.classification:
        export_from_ultralytics_to_hf_classification(
            dataset_dir=dataset_dir,
            repo_id=repo_id,
            label_names=label_names,
            merge_labels=merge_labels,
            is_openfoodfacts_dataset=is_openfoodfacts_dataset,
            openfoodfacts_flavor=openfoodfacts_flavor,
        )



import random
import string

from openfoodfacts.utils import get_logger

from ultralytics.engine.results import Results

logger = get_logger(__name__)


def format_annotation_results_from_ultralytics(
    results: Results,
    labels: list[str],
    label_mapping: dict[str, str] | None = None,
) -> list[dict]:
    annotation_results = []
    orig_height, orig_width = results.orig_shape
    boxes = results.boxes
    classes = boxes.cls.tolist()
    scores = boxes.conf.tolist()
    for i, xyxyn in enumerate(boxes.xyxyn):
        # Boxes found.
        if len(xyxyn) > 0:
            xyxyn = xyxyn.tolist()
            x1 = xyxyn[0] * 100
            y1 = xyxyn[1] * 100
            x2 = xyxyn[2] * 100
            y2 = xyxyn[3] * 100
            width = x2 - x1
            height = y2 - y1
            label_id = int(classes[i])
            label_name = labels[label_id]
            # Undocumented Label Studio feature, but we can add the individual bounding box score:
            # https://github.com/HumanSignal/label-studio/issues/2944#issuecomment-1249943187
            # This allows to:
            # - display each individual bounding score
            # - find tasks for which the model is not-confident using Labelr (to be implemented)
            score = scores[i]
            if label_mapping:
                label_name = label_mapping.get(label_name, label_name)
            annotation_results.append(
                {
                    "id": generate_id(),
                    "type": "rectanglelabels",
                    "from_name": "label",
                    "to_name": "image",
                    "original_width": orig_width,
                    "original_height": orig_height,
                    "image_rotation": 0,
                    "score": score,
                    "value": {
                        "rotation": 0,
                        "x": x1,
                        "y": y1,
                        "width": width,
                        "height": height,
                        "rectanglelabels": [label_name],
                    },
                },
            )
    return annotation_results


def generate_id(length: int = 10) -> str:
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))



from pydantic import BaseModel, Field


class SampleMeta(BaseModel):
    barcode: str | None = Field(
        ..., description="The barcode of the product, if applicable"
    )
    off_image_id: str | None = Field(
        ...,
        description="The Open Food Facts image ID associated with the image, if applicable",
    )
    image_url: str | None = Field(
        ..., description="The URL of the image, if applicable"
    )



COLORS = [
    "blue",
    "green",
    "yellow",
    "red",
    "purple",
    "orange",
    "pink",
    "brown",
    "gray",
    "black",
    "white",
]


def create_object_detection_label_config(labels_names: list[str]) -> str:
    """Create a Label Studio label configuration for object detection tasks.

    The format is the following:
    ```xml
    <View>
    <Image name="image" value="$image_url"/>
    <RectangleLabels name="label" toName="image">
    <Label value="nutrition-table" background="green"/>
        <Label value="nutrition-table-small" background="blue"/>
        <Label value="nutrition-table-small-energy" background="yellow"/>
        <Label value="nutrition-table-text" background="red"/>
    </RectangleLabels>
    </View>
    ```
    """
    if len(labels_names) > len(COLORS):
        raise ValueError(
            f"Too many labels ({len(labels_names)}) for the available colors ({len(COLORS)})."
        )
    labels_xml = "\n".join(
        f'    <Label value="{label}" background="{color}"/>'
        for label, color in zip(labels_names, COLORS[: len(labels_names)])
    )
    return f"""<View>
<Image name="image" value="$image_url"/>
<RectangleLabels name="label" toName="image">
{labels_xml}
</RectangleLabels>
</View>"""



import datetime

import typer

app = typer.Typer(no_args_is_help=True)


AVAILABLE_OBJECT_DETECTION_MODELS = [
    "yolov8n.pt",
    "yolov8s.pt",
    "yolov8m.pt",
    "yolov8l.pt",
    "yolov8x.pt",
    "yolov9t.pt",
    "yolov9s.pt",
    "yolov9m.pt",
    "yolov9c.pt",
    "yolov9e.pt",
    "yolov10n.pt",
    "yolov10s.pt",
    "yolov10m.pt",
    "yolov10b.pt",
    "yolov10l.pt",
    "yolov10x.pt",
    "yolo11n.pt",
    "yolo11s.pt",
    "yolo11m.pt",
    "yolo11l.pt",
    "yolo11x.pt",
    "yolo12n.pt",
    "yolo12s.pt",
    "yolo12m.pt",
    "yolo12l.pt",
    "yolo12x.pt",
]


@app.command()
def train_object_detection(
    wandb_project: str = typer.Option(
        "train-yolo", help="The Weights & Biases project name."
    ),
    wandb_api_key: str = typer.Option(..., envvar="WANDB_API_KEY"),
    hf_token: str = typer.Option(
        ...,
        help="The Hugging Face token, used to push the trained model to Hugging Face Hub.",
    ),
    run_name: str = typer.Option(..., help="A name for the training run."),
    add_date_to_run_name: bool = typer.Option(
        True, help="Whether to append the date to the run name."
    ),
    hf_repo_id: str = typer.Option(
        ..., help="The Hugging Face dataset repository ID to use to train."
    ),
    hf_trained_model_repo_id: str = typer.Option(
        ..., help="The Hugging Face repository ID where to push the trained model."
    ),
    epochs: int = typer.Option(100, help="Number of training epochs."),
    imgsz: int = typer.Option(640, help="Size of the image during training."),
    batch: int = typer.Option(64, help="Batch size for training."),
    model_name: str = typer.Option(
        "yolov8n.pt",
        help="The YOLO model variant to use for training. "
        "This should be a valid Ultralytics model name.",
    ),
):
    """Train an object detection model."""

    if model_name not in AVAILABLE_OBJECT_DETECTION_MODELS:
        raise typer.BadParameter(
            f"Invalid model name '{model_name}'. Available models are: {', '.join(AVAILABLE_OBJECT_DETECTION_MODELS)}"
        )

    datestamp = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")

    if add_date_to_run_name:
        run_name = f"{run_name}-{datestamp}"

    env_variables = {
        "HF_REPO_ID": hf_repo_id,
        "HF_TRAINED_MODEL_REPO_ID": hf_trained_model_repo_id,
        "HF_TOKEN": hf_token,
        "WANDB_PROJECT": wandb_project,
        "RUN_NAME": run_name,
        "WANDB_API_KEY": wandb_api_key,
        "EPOCHS": str(epochs),
        "IMGSZ": str(imgsz),
        "BATCH_SIZE": str(batch),
        "USE_AWS_IMAGE_CACHE": "False",
        "YOLO_MODEL_NAME": model_name,
    }

    job_name = f"train-yolo-job-{run_name}"
    if not add_date_to_run_name:
        # Ensure job name is unique by adding a datestamp if date is not added to run name
        job_name = f"{job_name}-{datestamp}"

    job = launch_job(
        job_name=job_name,
        container_image_uri="europe-west9-docker.pkg.dev/robotoff/gcf-artifacts/train-yolo",
        env_variables=env_variables,
    )
    typer.echo("Job launched")
    typer.echo(job)


def launch_job(
    job_name: str = typer.Argument(
        ...,
        help="The name of the Google Batch job that will be created. "
        "It needs to be unique for each project and region pair.",
    ),
    container_image_uri: str = typer.Argument(
        ..., help="The URI of the container image that will be run as part of the job."
    ),
    commands: str | None = None,
    env_variables: dict[str, str] | None = None,
    entrypoint: str | None = None,
    cpu_milli: int = 4000,  # in milli-CPU units (4000 = 4 CPUs). This means the task requires 4 whole CPUs.
    memory_mib: int = 16000,  # Make sure to have enough memory for the 2GB of shared memory set below.
    boot_disk_mib: int = 100000,
    max_retry_count: int = 1,
    max_run_duration: str = "86400s",  # 24 hours
    task_count: int = 1,
    accelerators_type: str = "nvidia-tesla-t4",
    machine_type: str = "n1-standard-8",
    google_project_id: str = "robotoff",
    accelerators_count: int = 1,
    region: str = "europe-west4",
    install_gpu_drivers: bool = True,
):
    """This method creates a Batch Job on GCP.

    Sources:
    * https://github.com/GoogleCloudPlatform/python-docs-samples/tree/main/batch/create
    * https://cloud.google.com/python/docs/reference/batch/latest/google.cloud.batch_v1.types  # noqa

    :param google_batch_launch_config: Config to run a job on Google Batch.
    :param batch_job_config: Config to run a specific job on Google Batch.
    :return: Batch job information.

    Returns:
        Batch job information.
    """
    from google.cloud import batch_v1

    client = batch_v1.BatchServiceClient()

    # Define what will be done as part of the job.
    runnable = batch_v1.Runnable()
    runnable.container = batch_v1.Runnable.Container()
    runnable.container.image_uri = container_image_uri
    runnable.container.entrypoint = entrypoint  # type: ignore
    # By default, /dev/shm is 64MB which is not enough for Pytorch
    runnable.container.options = "--shm-size=2048m"
    runnable.container.commands = commands

    # Jobs can be divided into tasks. In this case, we have only one task.
    task = batch_v1.TaskSpec()
    task.runnables = [runnable]

    # Environment variables.
    envable = batch_v1.Environment()
    envable.variables = env_variables or {}
    task.environment = envable

    # We can specify what resources are requested by each task.
    resources = batch_v1.ComputeResource()
    resources.cpu_milli = cpu_milli
    resources.memory_mib = memory_mib
    resources.boot_disk_mib = boot_disk_mib  # type: ignore
    task.compute_resource = resources

    task.max_retry_count = max_retry_count
    task.max_run_duration = max_run_duration  # type: ignore

    # Tasks are grouped inside a job using TaskGroups.
    group = batch_v1.TaskGroup()
    group.task_count = task_count  # type: ignore
    group.task_spec = task

    # Policies are used to define on what kind of virtual machines the tasks
    # will run on.
    policy = batch_v1.AllocationPolicy.InstancePolicy()
    # See list of machine types here:
    # https://docs.cloud.google.com/compute/docs/gpus#t4-gpus
    policy.machine_type = machine_type

    accelerator = batch_v1.AllocationPolicy.Accelerator()
    accelerator.type_ = accelerators_type
    accelerator.count = accelerators_count

    policy.accelerators = [accelerator]
    instances = batch_v1.AllocationPolicy.InstancePolicyOrTemplate()
    instances.policy = policy
    instances.install_gpu_drivers = install_gpu_drivers
    allocation_policy = batch_v1.AllocationPolicy()
    allocation_policy.instances = [instances]

    job = batch_v1.Job()
    job.task_groups = [group]
    job.allocation_policy = allocation_policy
    # We use Cloud Logging as it's an out of the box available option
    job.logs_policy = batch_v1.LogsPolicy()
    job.logs_policy.destination = batch_v1.LogsPolicy.Destination.CLOUD_LOGGING  # type: ignore

    create_request = batch_v1.CreateJobRequest()
    create_request.job = job
    create_request.job_id = job_name
    # The job's parent is the region in which the job will run
    create_request.parent = f"projects/{google_project_id}/locations/{region}"

    return client.create_job(create_request)



import typing
from collections.abc import Iterator
from pathlib import Path

import datasets
import orjson
from PIL import Image
from pydantic import BaseModel, Field

from labelr.sample.common import SampleMeta
from labelr.utils import download_image


class LLMImageExtractionSample(BaseModel):
    class Config:
        # required to allow PIL Image type
        arbitrary_types_allowed = True

    image_id: str = Field(
        ...,
        description="unique ID for the image. For Open Food Facts images, it follows the "
        "format `barcode:imgid`",
    )
    image: Image.Image = Field(..., description="Image to extract information from")
    output: str | None = Field(..., description="Expected response of the LLM")
    meta: SampleMeta = Field(..., description="Metadata associated with the sample")


HF_DS_LLM_IMAGE_EXTRACTION_FEATURES = datasets.Features(
    {
        "image_id": datasets.Value("string"),
        "image": datasets.features.Image(),
        "output": datasets.features.Value("string"),
        "meta": {
            "barcode": datasets.Value("string"),
            "off_image_id": datasets.Value("string"),
            "image_url": datasets.Value("string"),
        },
    }
)


def load_llm_image_extraction_dataset_from_jsonl(
    dataset_path: Path, **kwargs
) -> Iterator[LLMImageExtractionSample]:
    """Load a Hugging Face dataset for LLM image extraction from a JSONL file.

    Args:
        dataset_path (Path): Path to the JSONL dataset file.
        **kwargs: Additional keyword arguments to pass to the image downloader.
    Yields:
        Iterator[LLMImageExtractionSample]: Iterator of LLM image extraction
            samples.
    """
    with dataset_path.open("r") as f:
        for line in f:
            item = orjson.loads(line)
            image_id = item["image_id"]
            image_url = item["image_url"]
            image = typing.cast(Image.Image, download_image(image_url, **kwargs))
            barcode = item.pop("barcode", None)
            off_image_id = item.pop("off_image_id", None)
            output = item.pop("output", None)
            meta = SampleMeta(
                barcode=barcode,
                off_image_id=off_image_id,
                image_url=image_url,
            )
            sample = LLMImageExtractionSample(
                image_id=image_id,
                image=image,
                output=output,
                meta=meta,
            )
            yield sample



import typer


def check_label_studio_api_key(api_key: str | None):
    if not api_key:
        raise typer.BadParameter(
            "Label Studio API key not provided. Please provide it with the "
            "--api-key option or set the LABELR_LABEL_STUDIO_API_KEY environment variable."
        )



from pathlib import Path
from typing import Annotated

import typer
from openfoodfacts.utils import get_logger

from . import typer_description
from ...config import config, check_required_field
from .common import check_label_studio_api_key

app = typer.Typer(no_args_is_help=True)

logger = get_logger(__name__)


@app.command()
def add(
    tag: Annotated[
        str,
        typer.Option(help="Value of the tag to add"),
    ],
    task_id_file: Annotated[
        Path,
        typer.Option(
            help="Path of a text file containing IDs of samples",
            exists=True,
            dir_okay=False,
            file_okay=True,
            readable=True,
        ),
    ],
    project_id: Annotated[
        int | None, typer.Option(help=typer_description.LABEL_STUDIO_PROJECT_ID)
    ] = config.label_studio_project_id,
    api_key: Annotated[
        str | None, typer.Option(help=typer_description.LABEL_STUDIO_API_KEY)
    ] = config.label_studio_api_key,
    label_studio_url: Annotated[
        str, typer.Option(help=typer_description.LABEL_STUDIO_URL)
    ] = config.label_studio_url,
):
    """Add a tag to Label Studio tasks.

    Tags are used to filter tasks on Label Studio using arbitrary criteria.
    A plain text file containing IDs of Label Studio tasks (one per line) must be
    provided with --task-id-file. The provided tag will be added in the `tags` data
    field for each task.
    """

    from label_studio_sdk.client import LabelStudio

    check_label_studio_api_key(api_key)
    check_required_field("--project-id", project_id)

    typer.echo(
        f"Adding tag '{tag}' to tasks with IDs read from file '{task_id_file}'",
        err=True,
    )
    ls = LabelStudio(base_url=label_studio_url, api_key=api_key)
    updated = 0

    with task_id_file.open("r") as f:
        for line in f:
            task_id = line.strip()
            if not task_id:
                continue
            task = ls.tasks.get(task_id)

            if task:
                data = task.data
                data.setdefault("tags", [])
                if not isinstance(data["tags"], list):
                    raise ValueError(
                        f"tags field for task '{task.id}' is not a list: '{data['tags']}'"
                    )
                if tag in data["tags"]:
                    continue
                task.data["tags"].append(tag)
                ls.tasks.update(task_id, data=task.data)
                updated += 1
            else:
                typer.echo(f"Task '{task.id}' not found", err=True)

    typer.echo(f"Number of tasks updated: {updated}", err=True)



import datasets

HF_DS_CLASSIFICATION_FEATURES = datasets.Features(
    {
        "image_id": datasets.Value("string"),
        "image": datasets.features.Image(),
        "width": datasets.Value("int64"),
        "height": datasets.Value("int64"),
        "meta": {
            "barcode": datasets.Value("string"),
            "off_image_id": datasets.Value("string"),
            "image_url": datasets.Value("string"),
        },
        "category_id": datasets.Value("int64"),
        "category_name": datasets.Value("string"),
    }
)



import typing
from collections import defaultdict
from pathlib import Path

import imagehash
import tqdm
from label_studio_sdk.client import LabelStudio
from openfoodfacts.types import JSONType
from openfoodfacts.utils import ImageDownloadItem, get_image_from_url, get_logger
from PIL import Image

logger = get_logger(__name__)


def check_ls_dataset(
    ls: LabelStudio,
    project_id: int,
    view_id: int | None = None,
    delete_missing_images: bool = False,
    delete_duplicate_images: bool = False,
):
    """Perform sanity checks of a Label Studio dataset.

    This function checks for:
    - Tasks with missing images (404)
    - Duplicate images based on perceptual hash (pHash)
    - Tasks with multiple annotations

    This function doesn't perform any modifications to the dataset, except
    optionally deleting tasks with missing images if `delete_missing_images`
    is set to True.

    Args:
        ls (LabelStudio): Label Studio client instance.
        project_id (int): ID of the Label Studio project to check.
        view_id (int): ID of the Label Studio view to check. If None, no
            filtering is done.
        delete_missing_images (bool): Whether to delete tasks with missing
            images.
        delete_duplicate_images (bool): Whether to delete tasks with duplicate
            images. If one task has annotations and the other doesn't, the task
            with annotations will be kept. Otherwise, the most recent task will
            be kept.
    """
    skipped = 0
    not_annotated = 0
    annotated = 0
    deleted = 0
    multiple_annotations = 0
    hash_map = defaultdict(list)
    for task in tqdm.tqdm(
        ls.tasks.list(project=project_id, fields="all", view=view_id), desc="tasks"
    ):
        annotations = typing.cast(list[JSONType], task.annotations)

        if len(annotations) == 0:
            not_annotated += 1
            continue
        elif len(annotations) > 1:
            logger.warning("Task has multiple annotations: %s", task.id)
            multiple_annotations += 1
            continue

        annotation = annotations[0]

        if annotation["was_cancelled"]:
            skipped += 1

        annotated += 1
        image_url = task.data["image_url"]
        image_struct = typing.cast(
            ImageDownloadItem,
            get_image_from_url(image_url, return_struct=True, error_raise=False),
        )

        if image_struct.response.status_code == 404:
            logger.warning("Image not found (404): %s", image_url)

            if delete_missing_images:
                ls.tasks.delete(task.id)
                deleted += 1
                logger.info("Deleted task with missing image: %s", task.id)
            continue

        if image_struct.image is None:
            logger.warning("Could not open image: %s", image_url)
            continue

        image_hash = str(imagehash.phash(image_struct.image))
        hash_map[image_hash].append(task.id)

    for image_hash, task_ids in hash_map.items():
        if len(task_ids) > 1:
            logger.warning("Duplicate images: %s", task_ids)
            if delete_duplicate_images:
                tasks = [ls.tasks.get(id=task_id) for task_id in task_ids]
                # We sort the tasks by the number of annotations, so that we keep the
                # one with at least one annotation.
                for task in sorted(tasks, key=lambda x: len(x.annotations) > 0)[:-1]:
                    logger.info("Deleting duplicate task: %s", task.id)
                    ls.tasks.delete(task.id)
                    deleted += 1

    logger.info(
        "Tasks - annotated: %d, skipped: %d, not annotated: %d, multiple annotations: %d",
        annotated,
        skipped,
        not_annotated,
        multiple_annotations,
    )
    logger.info("Deleted tasks with missing images: %d", deleted)


def check_local_dataset(dataset_dir: Path, remove: bool = False):
    hash_map = defaultdict(list)
    for path in tqdm.tqdm(dataset_dir.glob("**/*.jpg"), desc="images"):
        if path.is_file() and path.suffix in [
            ".jpg",
            ".jpeg",
            ".png",
            ".webp",
            ".bmp",
            ".tiff",
            ".gif",
        ]:
            image = Image.open(path)
            image_hash = str(imagehash.phash(image))
            logger.debug("Image hash: %s", image_hash)
            hash_map[image_hash].append(path)

    duplicated = 0
    to_remove = []
    for image_hash, image_paths in hash_map.items():
        if len(image_paths) > 1:
            logger.warning(
                "Duplicate images: %s",
                [str(x.relative_to(dataset_dir)) for x in image_paths],
            )
            duplicated += 1
            to_remove.append(image_paths[0])

    logger.info("Total duplicated groups: %d", duplicated)

    if remove and to_remove:
        for path in to_remove:
            logger.info("Removing: %s", str(path))
            path.unlink()



from pathlib import Path
from typing import Annotated

import requests
import typer

app = typer.Typer(no_args_is_help=True)


DEFAULT_DIRECTUS_URL = "http://localhost:8055"


def _list_endpoint_iter(
    url: str,
    session: requests.Session,
    page_size: int,
    method: str = "GET",
    list_field: str | None = "data",
    **kwargs,
):
    """Iterate over paginated Directus endpoint.

    Args:
        url (str): URL of the Directus endpoint.
        session (requests.Session): Requests session to use for making HTTP
            requests.
        page_size (int): Number of items to fetch per page.
        method (str, optional): HTTP method to use. Defaults to "GET".
        list_field (str | None, optional): Field in the response JSON that
            contains the list of items. If None, the entire response is used as
            the list. Defaults to "data".
        **kwargs: Additional keyword arguments to pass to the requests method.
    Yields:
        dict: Items from the Directus endpoint.
    """
    page = 0
    next_page = True
    params = kwargs.pop("params", {})

    while next_page:
        params["offset"] = page * page_size
        params["limit"] = page_size
        r = session.request(method=method, url=url, params=params, **kwargs)
        r.raise_for_status()
        response = r.json()
        items = response[list_field] if list_field else response
        if len(items) > 0:
            yield from items
        else:
            next_page = False
        page += 1


def iter_items(
    collection_name: str,
    url: str,
    session: requests.Session,
    page_size: int = 50,
    **kwargs,
):
    """Iterate over items in a Directus collection.

    Args:
        collection_name (str): Name of the Directus collection.
        url (str): Base URL of the Directus server.
        session (requests.Session): Requests session to use for making HTTP
            requests.
        page_size (int, optional): Number of items to fetch per page. Defaults
            to 50.
        **kwargs: Additional keyword arguments to pass to the requests method.
    Yields:
        dict: Items from the Directus collection.
    """
    yield from _list_endpoint_iter(
        url=f"{url}/items/{collection_name}",
        session=session,
        page_size=page_size,
        **kwargs,
    )


@app.command()
def upload_data(
    dataset_path: Annotated[
        Path,
        typer.Option(
            help="Path to the dataset JSONL file to upload from.",
            file_okay=True,
            dir_okay=False,
            readable=True,
        ),
    ],
    collection: Annotated[
        str, typer.Option(help="Name of the collection to upload the items to.")
    ],
    directus_url: Annotated[
        str,
        typer.Option(
            help="Base URL of the Directus server.",
        ),
    ] = DEFAULT_DIRECTUS_URL,
):
    """Upload data to a Directus collection."""
    import orjson
    import requests
    import tqdm

    session = requests.Session()

    with dataset_path.open("r") as f:
        for item in tqdm.tqdm(map(orjson.loads, f), desc="items"):
            r = session.post(
                f"{directus_url}/items/{collection}",
                json=item,
            )
            print(r.json())
            r.raise_for_status()


@app.command()
def update_items(
    collection: Annotated[
        str, typer.Option(help="Name of the collection to upload the items to.")
    ],
    directus_url: Annotated[
        str,
        typer.Option(
            help="Base URL of the Directus server.",
        ),
    ] = DEFAULT_DIRECTUS_URL,
    sort: Annotated[
        str | None,
        typer.Option(help="The field to sort items by, defaults to None (no sorting)."),
    ] = None,
    skip: Annotated[
        int, typer.Option(help="Number of items to skip, defaults to 0.")
    ] = 0,
):
    """Update items in a Directus collection.

    **Warning**: This command requires you to implement the processing
    function inside the command. It is provided as a template for batch
    updating items in a Directus collection.
    """
    import requests
    import tqdm

    session = requests.Session()

    params = {} if sort is None else {"sort[]": sort}
    for i, item in tqdm.tqdm(
        enumerate(
            iter_items(
                collection_name=collection,
                url=directus_url,
                session=session,
                params=params,
            )
        )
    ):
        if i < skip:
            typer.echo(f"Skipping item {i}")
            continue

        item_id = item["id"]
        # Implement your processing function here
        # It should return a dict with the fields to update only
        # If no update is needed, it should return None
        patch_item = None

        if patch_item is not None:
            r = session.patch(
                f"{directus_url}/items/{collection}/{item_id}",
                json=patch_item,
            )
            r.raise_for_status()


@app.command()
def export_data(
    output_path: Annotated[
        Path, typer.Option(help="Path to the file to export to.", allow_dash=True)
    ],
    collection: Annotated[
        str, typer.Option(help="Name of the collection to upload the items to.")
    ],
    directus_url: Annotated[
        str,
        typer.Option(
            help="Base URL of the Directus server.",
        ),
    ] = DEFAULT_DIRECTUS_URL,
):
    """Export a directus collection to a JSONL file."""
    import sys

    import orjson
    import requests
    import tqdm

    session = requests.Session()

    f = sys.stdout if output_path.as_posix() == "-" else output_path.open("w")
    with f:
        for item in tqdm.tqdm(
            iter_items(
                collection_name=collection,
                url=directus_url,
                session=session,
            )
        ):
            f.write(orjson.dumps(item).decode("utf-8") + "\n")



LABEL_STUDIO_API_KEY = """API Key to authenticate to the Label Studio server. Can also be set with the LABELR_LABEL_STUDIO_API_KEY environment variable."""
LABEL_STUDIO_URL = """URL of the Label Studio server. Can also be set with the LABELR_LABEL_STUDIO_URL environment variable."""
LABEL_STUDIO_PROJECT_ID = "ID of the Label Studio project"
LABEL_STUDIO_VIEW_ID = "ID of the Label Studio view, used to filter tasks. If not provided, all tasks in the project are processed."



from pathlib import Path
from typing import Annotated

import typer

app = typer.Typer(no_args_is_help=True)


@app.command()
def show_hf_sample(
    repo_id: Annotated[
        str,
        typer.Argument(
            ...,
            help="Hugging Face Datasets repo ID. The revision can be specified by "
            "appending `@<revision>` to the repo ID.",
        ),
    ],
    image_id: Annotated[
        str,
        typer.Argument(
            ...,
            help="ID of the image associated with the sample to display (field: `image_id`)",
        ),
    ],
    output_image_path: Annotated[
        Path | None,
        typer.Option(help="Path to save the sample image (optional)", exists=False),
    ] = None,
):
    """Display a sample from a Hugging Face Datasets repository by image ID."""
    from labelr.utils import parse_hf_repo_id

    repo_id, revision = parse_hf_repo_id(repo_id)

    from datasets import load_dataset

    ds = load_dataset(repo_id, revision=revision)

    sample = None
    for split in ds.keys():
        samples = ds[split].filter(lambda x: x == image_id, input_columns="image_id")
        if len(samples) > 0:
            sample = samples[0]
            break
    if sample is None:
        typer.echo(f"Sample with image ID {image_id} not found in dataset {repo_id}")
        raise typer.Exit(code=1)

    else:
        for key, value in sample.items():
            typer.echo(f"{key}: {value}")

        if output_image_path is not None:
            image = sample["image"]
            image.save(output_image_path)
            typer.echo(f"Image saved to {output_image_path}")
